import copy
import datetime
import os
import time
from typing import Literal
import pandas as pd
import plotly.express as px
import tabulate
from wiliot_deployment_tools.ag.ut_defines import BRIDGE_ID, NFPKT, PAYLOAD, RSSI, LAT, LNG
from wiliot_deployment_tools.api.extended_api import GatewayType
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.gw_certificate.api_if.gw_capabilities import GWCapabilities
from wiliot_deployment_tools.gw_certificate.tests.static.coupling_defines import INCREMENTAL_STAGE_ADVA
from wiliot_deployment_tools.interface.ble_simulator import BLESimulator
from wiliot_deployment_tools.interface.if_defines import BRIDGES, DEFAULT_DELAY, LOCATION
from wiliot_deployment_tools.interface.uart_if import UARTInterface
from wiliot_deployment_tools.gw_certificate.tests.static.uplink_defines import *
from wiliot_deployment_tools.interface.mqtt import MqttClient, Serialization
from wiliot_deployment_tools.interface.pkt_generator import BrgPktGenerator, BrgPktGeneratorNetwork, TagPktGenerator
from wiliot_deployment_tools.gw_certificate.tests.static.generated_packet_table import TEST_COUPLING, TEST_UPLINK, TEST_UNIFIED, CouplingRunData, UplinkRunData, UnifiedRunData, SensorRunData
from wiliot_deployment_tools.gw_certificate.tests.generic import PassCriteria, PERFECT_SCORE, MINIMUM_SCORE, INCONCLUSIVE_MINIMUM, INIT_INCONCLUSIVE_MINIMUM, GenericTest, GenericStage
from wiliot_deployment_tools.interface.packet_error import PacketError
from wiliot_deployment_tools.gw_certificate.api_if.api_validation import MESSAGE_TYPES, validate_message
from wiliot_deployment_tools.gw_certificate.tests.static.generated_packet_table import CSV_NAME
import pkg_resources


# HELPER DEFINES
TABLE_SUFFIX = "Table"

# HELPER FUNCTIONS

def process_payload(packet:dict):
    payload = packet[PAYLOAD]
    payload = payload.upper()
    if len(payload) == 62:
        if payload[:4] == '1E16':
            payload = payload [4:]
    # big2little endian
    if payload[:4] == 'FCC6':
        payload = 'C6FC' + payload[4:]
    packet[PAYLOAD] = payload
    return packet


# TEST STAGES

class UplinkTestError(Exception):
    pass

class GenericUplinkStage(GenericStage):
    def __init__(self, mqttc:MqttClient, ble_sim:BLESimulator, gw_capabilities:GWCapabilities, stage_name,
                **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(stage_name=stage_name, **self.__dict__)
                
        # Clients
        self.mqttc = mqttc
        self.ble_sim = ble_sim

        # Packets list
        self.local_pkts = []
        self.mqtt_pkts = []
        
        # GW Capabilities
        self.gw_capabilities = gw_capabilities
        
        # Packet Error / Run data
        self.packet_error = PacketError()
        self.run_data = UplinkRunData
        
        # Unified stage
        self.run_data_unified = UnifiedRunData
    
    def prepare_stage(self, reset_ble_sim=True):
        super().prepare_stage()
        self.mqttc.flush_messages()
        if reset_ble_sim:
            self.ble_sim.set_sim_mode(True) 
        
    def fetch_mqtt_from_stage(self):
        mqtt_pkts = self.mqttc.get_all_tags_pkts()
        self.mqtt_pkts = list(map(lambda p: process_payload(p), mqtt_pkts))
    
    ## TODO - REWRITE
    def compare_local_mqtt(self):
        self.fetch_mqtt_from_stage()
        local_pkts_df = pd.DataFrame(self.local_pkts)
        mqtt_pkts_df = pd.DataFrame(self.mqtt_pkts)
        comparison = local_pkts_df

        if PAYLOAD not in mqtt_pkts_df.columns:
            mqtt_pkts_df[PAYLOAD] = ''
        received_pkts_df = pd.merge(local_pkts_df[PAYLOAD], mqtt_pkts_df[PAYLOAD], how='inner')
        
        received_pkts = set(received_pkts_df[PAYLOAD])

        self.pkts_received_count = pd.Series.count(received_pkts_df)
        unique_received_count = len(received_pkts)
        self.pkts_filtered_out_count = self.pkts_received_count - unique_received_count

        comparison[RECEIVED] = comparison[PAYLOAD].isin(received_pkts)
        comparison['pkt_id'] = comparison['payload'].apply(lambda x: x[-8:])
        self.comparison = comparison
                
    def generate_stage_report(self):
        """
        Generates report for the stage
        """
        self.compare_local_mqtt()
        report = []
        num_pkts_sent = len(self.comparison)
        num_pkts_received = self.comparison['received'].eq(True).sum()
        self.stage_pass = num_pkts_received / num_pkts_sent * PERFECT_SCORE
        self.stage_pass, self.error_summary = PassCriteria.calc_for_stage_uplink(self.stage_pass, self.stage_name)

        report.append(((f'Number of unique packets sent'), num_pkts_sent))
        report.append(((f'Number of unique packets received'), num_pkts_received))
        report.append(((f'Number of total packets received'), self.pkts_received_count))
        report.append(((f'Number of duplicates out of total'), self.pkts_filtered_out_count))
        self.add_to_stage_report(f'---Stage {self.stage_name} {PassCriteria.to_string(self.stage_pass)}, Running time {datetime.datetime.now() - self.start_time}')
        self.add_to_stage_report(tabulate.tabulate(pd.DataFrame(report), showindex=False))
        not_received = self.comparison[self.comparison[RECEIVED]==False][REPORT_COLUMNS]
        if len(not_received) > 0:
            self.add_to_stage_report('Packets not received:')
            self.add_to_stage_report(tabulate.tabulate(not_received, headers='keys', showindex=False))
        self.comparison.to_csv(self.csv_path)
        self.add_to_stage_report(f'Stage data saved - {self.csv_path}')
        debug_print(self.report)
        
        # Generate HTML
        table_html = self.template_engine.render_template('table.html', dataframe=self.comparison.to_html(table_id=self.stage_name + TABLE_SUFFIX),
                                                          table_id=self.stage_name + TABLE_SUFFIX)
        self.report_html = self.template_engine.render_template('stage.html', stage=self, 
                                                                stage_report=self.report.split('\n'), table=table_html)
        
        return self.report
    
# Unified Stages
class OneBrgUnifiedPacketStage(GenericUplinkStage):
# currently only one bridge is simulated. If we want to simulate more, then make sure that they have different parameters to send (rssi,nfpkt,latency,global pacing group)
    def __init__(self, **kwargs):
        self.stage_tooltip = "Simulates advertisements from a single bridge. Expects the gateway to scan & upload them"

        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)
        self.pkt_gen = BrgPktGenerator()

    def run(self):
        self.start_time = datetime.datetime.now()
        for duplication in UNIFIED_DUPLICATIONS:
            debug_print(f'Duplication {duplication}')
            for time_delay in UNIFIED_TIME_DELAYS:
                debug_print(f'Time Delay {time_delay}')
                run_data = self.run_data_unified.get_data(TEST_UNIFIED, duplication, time_delay, BRIDGES[0])
                data = run_data.data
                self.local_pkts.extend(run_data.expected_mqtt)
                for dup in range(duplication):
                    self.ble_sim.send_packet(data, duplicates=1, delay=time_delay)
            time.sleep(5)

class ThreeBrgUnifiedPacketStage(GenericUplinkStage):  

    def __init__(self, **kwargs):
        self.stage_tooltip = "Simulates advertisements from three bridges. Expects the gateway to scan & upload them"
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)
        self.brg_network = BrgPktGeneratorNetwork()
    
    def run(self):
        self.start_time = datetime.datetime.now()
        for duplication in UNIFIED_DUPLICATIONS:
            debug_print(f'Duplication {duplication}')
            for time_delay in UNIFIED_TIME_DELAYS:
                debug_print(f'Time Delay {time_delay}')
                # Construct packet list from data
                pkts = []
                for brg_idx in BRIDGES:
                    pkt = {}
                    run_data = self.run_data_unified.get_data(TEST_UNIFIED, duplication, time_delay, brg_idx)
                    pkt['data_packet'] = run_data.data
                    pkt['bridge_id'] = run_data.bridge_id
                    self.local_pkts.extend(run_data.expected_mqtt)
                    pkts.append(pkt)
                # Send packets
                self.ble_sim.send_brg_network_pkts_unified(pkts, duplication, delay=time_delay)
        time.sleep(5)



class SensorPacketStage(GenericUplinkStage):
    def __init__(self, **kwargs):
        self.stage_tooltip = ""

        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)
        self.pkt_gen = BrgPktGenerator()

    def run(self):
        self.start_time = datetime.datetime.now()
        run_data = SensorRunData()
        run_data = run_data.data
        for index, row in run_data.iterrows():
            data = row['data']
            si = row['si']
            expected_mqtt = eval(row['expected_mqtt'])
            self.local_pkts.extend(expected_mqtt)
            self.ble_sim.send_data_si_pair(data_packet=data, si_packet=si, duplicates=5)
        time.sleep(5)
        
    def compare_local_mqtt(self):
        self.fetch_mqtt_from_stage()
        local_pkts_df = pd.DataFrame(self.local_pkts)
        mqtt_pkts_df = pd.DataFrame(self.mqtt_pkts)
        comparison = local_pkts_df

        if not set(SHARED_COLUMNS) <= set(mqtt_pkts_df.columns):
            missing_columns = list(set(SHARED_COLUMNS) - set(mqtt_pkts_df.columns))
            for missing_column in missing_columns:
                if missing_column in OBJECT_COLUMNS:
                    mqtt_pkts_df[missing_column] = ''
                if missing_column in INT64_COLUMNS:
                    mqtt_pkts_df[missing_column] = 0
        received_pkts_df = pd.merge(local_pkts_df[SHARED_COLUMNS], mqtt_pkts_df[SHARED_COLUMNS], how='inner')
        
        received_pkts = set(received_pkts_df[PAYLOAD])

        self.pkts_received_count = pd.Series.count(received_pkts_df)
        unique_received_count = len(received_pkts)
        self.pkts_filtered_out_count = self.pkts_received_count - unique_received_count

        comparison[RECEIVED] = comparison[PAYLOAD].isin(received_pkts)
        comparison['pkt_id'] = comparison['payload'].apply(lambda x: x[-8:])
        self.comparison = comparison
        
    def generate_stage_report(self):
        self.compare_local_mqtt()
        print(self.comparison)
        report = []
        num_pkts_sent = len(self.comparison)
        num_pkts_received = self.comparison['received'].eq(True).sum()
        pkt_id_pairs = self.comparison.groupby('pkt_id').filter(lambda x: x['received'].all() and len(x) == 2)
        unique_pkt_ids = pkt_id_pairs['pkt_id'].unique()
        num_pairs = len(unique_pkt_ids)
        
        if num_pairs > 1:
            self.stage_pass = PERFECT_SCORE
        else:
            self.stage_pass = MINIMUM_SCORE
            
        self.stage_pass, self.error_summary = PassCriteria.calc_for_stage_uplink(self.stage_pass, self.stage_name)

        report.append(((f'Number of sensor packets sent'), num_pkts_sent / 2))
        report.append(((f'Number of sensor packets received correctly'), num_pairs))
        self.add_to_stage_report(f'---Stage {self.stage_name} {PassCriteria.to_string(self.stage_pass)}, Running time {datetime.datetime.now() - self.start_time}')
        self.add_to_stage_report(tabulate.tabulate(pd.DataFrame(report), showindex=False))
        not_received = self.comparison[self.comparison[RECEIVED]==False][REPORT_COLUMNS]
        if len(not_received) > 0:
            self.add_to_stage_report('Packets not received:')
            self.add_to_stage_report(tabulate.tabulate(not_received, headers='keys', showindex=False))
        self.comparison.to_csv(self.csv_path)
        self.add_to_stage_report(f'Stage data saved - {self.csv_path}')
        debug_print(self.report)
        
        # Generate HTML
        table_html = self.template_engine.render_template('table.html', dataframe=self.comparison.to_html(table_id=self.stage_name + TABLE_SUFFIX),
                                                          table_id=self.stage_name + TABLE_SUFFIX)
        self.report_html = self.template_engine.render_template('stage.html', stage=self, 
                                                                stage_report=self.report.split('\n'), table=table_html)
        
        return self.report
        
    
class ApiValidationStage(GenericUplinkStage):
    def __init__(self, **kwargs):
        self.stage_tooltip = "Validates the JSON structure of messages uploaded by the gateway in previous stages"
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)
    
    def prepare_stage(self):
        super().prepare_stage(reset_ble_sim=False)
        self.mqttc.flush_messages()

    def run(self):
        pass

    def generate_stage_report(self, **kwargs):
        report = []
        all_validations = []
        self.stage_pass = PERFECT_SCORE
                                
        for idx, message in enumerate(self.all_messages_in_test):
            message_body = message.body
            if len(message_body['packets']) == 0:
                continue
            validation = validate_message(MESSAGE_TYPES.DATA, message_body)
            errors = []
            for e in validation[1]:
                if e.message not in errors:
                    errors.append(e.message)
            all_validations.append({'valid':validation[0], 'errors': errors, 'message': message_body,})
            if not validation[0]:
                if 'Validation Errors:' not in report:
                    report.append('Validation Errors:')
                report.append(f'- Message (idx={idx}) Errors:')
                for e in errors:
                    report.append(e)
                self.stage_pass = MINIMUM_SCORE
                error_explainer = "API (JSON strcture) is invalid"
                if error_explainer not in self.error_summary:
                    self.error_summary = error_explainer


        # Set stage as FAIL if no messages were received:
        if len(self.all_messages_in_test) == 0:
            self.stage_pass = MINIMUM_SCORE
            self.error_summary = "No packets were received"

        self.add_to_stage_report(f'---Stage {self.stage_name} {PassCriteria.to_string(self.stage_pass)}')
        # Add all messages that failed to validate to report
        for line in report:
            self.add_to_stage_report(line)
        all_validations_df = pd.DataFrame(all_validations)
        all_validations_df.to_csv(self.csv_path)
        self.add_to_stage_report(f'Stage data saved - {self.csv_path}')
        debug_print(self.report)
        
        #Generate HTML
        table_html = self.template_engine.render_template('table.html', dataframe=all_validations_df.to_html(table_id=self.stage_name + TABLE_SUFFIX),
                                                          table_id=self.stage_name + TABLE_SUFFIX)
        self.report_html = self.template_engine.render_template('stage.html', stage=self, 
                                                                stage_report=self.report.split('\n'))
        return self.report

class SequentialSequenceIdStage(GenericUplinkStage):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        self.stage_tooltip = "Validates expected sequenceId in all packets"
        super().__init__(**self.__dict__, stage_name=type(self).__name__)
    
    def prepare_stage(self):
        super().prepare_stage(reset_ble_sim=False)
        self.mqttc.flush_messages()

    def run(self):
        pass

    def generate_stage_report(self, **kwargs):
        report = []
        self.stage_pass = PERFECT_SCORE
        required_sequenceId = None
        sequenceId_valid = True

        def is_sequenceId_incremental(idx, message):
            nonlocal required_sequenceId, sequenceId_valid
            packets = message['packets']

            # check that there is sequenceId in all packets
            packets_w_seqid = list(filter(lambda p: 'sequenceId' in p, packets))
            if len(packets_w_seqid) == 0:
                sequenceId_valid = False
                report.append(f'No sequenceId in message {idx}. Expected sequenceId in all packets')
                self.error_summary = self.error_summary + 'No SequenceId in packets.'
                return False
            
            # initialize the required sequenceId 
            if idx == 0:
                first_pkt = packets[0]
                required_sequenceId = first_pkt['sequenceId'] 
            
            # check that for every packet in message the sequenceId is incremental:
            for pkt in packets:
                pkt_sequenceId = pkt['sequenceId']
                if pkt_sequenceId != required_sequenceId:
                    if sequenceId_valid == True:
                        report.append(f'SequenceId is not incremental. Expected sequenceId is {required_sequenceId} but the packet sequenceId is {pkt_sequenceId}')
                        self.stage_pass = MINIMUM_SCORE
                        self.error_summary = self.error_summary + 'SequenceId is not incremental. '
                        sequenceId_valid = False
                    break
                required_sequenceId += 1

        # Set message type according to coupling, location
        for idx, message in enumerate(self.all_messages_in_test):
            message_body = message.body
            is_sequenceId_incremental(idx=idx, message=message_body)

        self.add_to_stage_report(f'---Stage {self.stage_name} {PassCriteria.to_string(self.stage_pass)}')
        self.add_to_stage_report(f"{'---SequenceId is incremental' if sequenceId_valid else '---SequenceId is NOT incremental'}")
        for line in report:
            self.add_to_stage_report(line)
        debug_print(self.report)
        
        #Generate HTML
        self.report_html = self.template_engine.render_template('stage.html', stage=self, 
                                                                stage_report=self.report.split('\n'))
        return self.report

class AliasBridgeIDStage(GenericUplinkStage):
    def __init__(self, **kwargs):
        self.stage_tooltip = "Validates the expected alias bridge ID (For unified SideInfo packets)"
        # Data extracted from the test csv
        self.all_test_payloads = None
        self.alias_bridge_id_df = None
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)
    
    def prepare_stage(self):
        super().prepare_stage(reset_ble_sim=False)
        self.mqttc.flush_messages()

    def run(self):
        pass

    def get_data_from_test_csv(self):
        relative_path = 'static/' + CSV_NAME
        csv_path = pkg_resources.resource_filename(__name__, relative_path)
        df = pd.read_csv(csv_path)

        # Store all test payloads
        all_payloads = df['data'].str[12:]
        self.all_test_payloads = all_payloads.tolist()

        def _parser(row, desired:Literal['adva', 'without_adva']):
            if desired == 'adva':
                output_string = row.at['data'][:12]
            elif desired == 'without_adva':
                output_string = row.at['data'][12:]
            else:
                raise ValueError
            return output_string

        # Create data set for alias bridge verification
        alias_bridge_id_df = df[df['test'] == 'unified'].copy()
        alias_bridge_id_df['payload'] = alias_bridge_id_df.apply(lambda row: _parser(row, 'without_adva'), axis=1)
        # .apply(lambda row: self._generate_interferers_data(), axis=1)
        alias_bridge_id_df['alias_bridge_id'] = alias_bridge_id_df.apply(lambda row: _parser(row, 'adva'), axis=1)
        # Convert bridge_id to little endian:
        alias_bridge_id_df['alias_bridge_id'] = alias_bridge_id_df['alias_bridge_id'].apply(lambda x: ''.join(format(byte, '02X') for byte in bytes.fromhex(x)[::-1]))
        self.alias_bridge_id_df = alias_bridge_id_df
   
    def generate_stage_report(self, **kwargs):
        report = []
        self.stage_pass = PERFECT_SCORE
        self.get_data_from_test_csv()
        aliasBridgeId_valid = True
            
        def filter_non_test_packets(message):
            packets = message['packets']
            filtered_pkts = []
            for pkt in packets:
                pkt = process_payload(pkt)
                payload = pkt['payload']
                if any(payload in test_payload for test_payload in self.all_test_payloads):
                    filtered_pkts.append(pkt)
            message['packets'] = filtered_pkts
                    
        def is_alias_bridge_id_valid(message): 
            nonlocal aliasBridgeId_valid
            packets = message['packets']

            for pkt in packets:
                if 'aliasBridgeId' in pkt: 
                    pkt_payload = pkt['payload']
                    pkt_alias_bridge_id = pkt['aliasBridgeId']
                    validation_data = self.alias_bridge_id_df[self.alias_bridge_id_df['payload'].str.contains(pkt_payload, case=False)] 
                    required_bridge_id = validation_data['alias_bridge_id'].iat[0]
                    if required_bridge_id != pkt_alias_bridge_id.upper():
                        report.append(f"Alias bridge ID of the packet does not match. The required alias bridge ID is {required_bridge_id} but the packet alias bridge ID is {pkt_alias_bridge_id}")
                        self.stage_pass = MINIMUM_SCORE
                        self.error_summary = "aliasBridgeId doesn't match the expected one of a packet. "
                        aliasBridgeId_valid = False 

        # Set message type according to coupling, location
        for idx, message in enumerate(self.all_messages_in_test):
            message_body = message.body
            filter_non_test_packets(message_body)
            if len(message_body['packets']) == 0:
                continue
            is_alias_bridge_id_valid(message=message_body)
        # Set stage as FAIL if no messages were received:
        if len(self.all_messages_in_test) == 0:
            self.stage_pass = MINIMUM_SCORE
            self.error_summary = "No packets were received"

        self.add_to_stage_report(f'---Stage {self.stage_name} {PassCriteria.to_string(self.stage_pass)}')
        self.add_to_stage_report(f"{'---Alias bridge ID is valid' if aliasBridgeId_valid else '---Alias bridge ID is NOT valid'}")
        for line in report:
            self.add_to_stage_report(line)
        # Add all messages that failed to validate to report
        debug_print(self.report)
        
        #Generate HTML
        self.report_html = self.template_engine.render_template('stage.html', stage=self, 
                                                                stage_report=self.report.split('\n'))
        return self.report



class GeolocationStage(GenericUplinkStage):
    def __init__(self, **kwargs):
        self.stage_tooltip = "Checks if lat/lng were uploaded under 'location' (optional JSON key) in the uploaded data messages"
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)
        self.graph_html_path = os.path.join(self.test_dir, f'{self.stage_name}.html')

    
    def prepare_stage(self):
        super().prepare_stage(reset_ble_sim=False)
        self.mqttc.flush_messages()

    def run(self):
        pass
    
    def generate_stage_report(self, **kwargs):
        locations_list = []
        locations_df = pd.DataFrame()
        self.stage_pass = MINIMUM_SCORE
        self.error_summary = "No coordinates were uploaded. "

        # Set message type according to coupling, location
        for message in self.all_messages_in_test:
            message = message.body
            timestamp = message[TIMESTAMP]
            if LOCATION in message.keys():
                loc = message[LOCATION]
                loc.update({TIMESTAMP:timestamp})
                locations_list.append(loc)
        num_unique_locs = 0
        if len(locations_list) > 0:
            self.stage_pass = PERFECT_SCORE
            self.error_summary = ''
            locations_df = pd.DataFrame(locations_list)
            num_unique_locs = locations_df[['lat', 'lng']].drop_duplicates().shape[0]
            fig = px.scatter_mapbox(locations_df, lat=LAT, lon=LNG, color='timestamp', zoom=10)
            fig.update(layout_coloraxis_showscale=False)
            fig.update_layout(scattermode="group", scattergap=0.95, mapbox_style="open-street-map")

        self.add_to_stage_report(f'---Stage {self.stage_name} {PassCriteria.to_string(self.stage_pass)}')
        self.add_to_stage_report(f'Number of unique locations received: {num_unique_locs}')
        # Export all stage data
        locations_df.to_csv(self.csv_path)
        self.add_to_stage_report(f'Stage data saved - {self.csv_path}')
        if num_unique_locs > 0:
            fig.write_html(self.graph_html_path)
        debug_print(self.report)
        
        #Generate HTML
        graph_div = fig.to_html(full_html=False, include_plotlyjs='cdn') if num_unique_locs > 0 else "No graph to display"
        self.report_html = self.template_engine.render_template('stage.html', stage=self, 
                                                                stage_report=self.report.split('\n'), graph = graph_div)
        return self.report


# TEST CLASS
UNCOUPLED_STAGES = [OneBrgUnifiedPacketStage, ThreeBrgUnifiedPacketStage, SequentialSequenceIdStage, AliasBridgeIDStage, SensorPacketStage] 

class UplinkTest(GenericTest):
    def __init__(self, **kwargs):
        self.test_tooltip = "Stages related to gateway BLE scans & MQTT data uploads"
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, test_name=type(self).__name__)
        self.all_messages_in_test = []
        stages = UNCOUPLED_STAGES
        if self.mqttc.get_serialization() == Serialization.JSON:
            stages = stages + ApiValidationStage
        if self.gw_capabilities.geoLocationSupport:
            stages.append(GeolocationStage)
        self.stages = [stage(**self.__dict__) for stage in stages]
        

    def run(self):
        super().run()
        self.test_pass = PERFECT_SCORE
        for stage in self.stages:
            stage.prepare_stage()
            stage.run()
            self.add_to_test_report(stage.generate_stage_report())
            self.test_pass = PassCriteria.calc_for_test(self.test_pass, stage)
            self.all_messages_in_test.extend(self.mqttc.get_all_messages_from_topic('data'))
    



from argparse import ArgumentParser
import datetime
import os
import time
from wiliot_core.utils.utils import WiliotDir
from wiliot_deployment_tools.interface.ble_sniffer import BLESniffer
from wiliot_deployment_tools.interface.uart_if import UARTInterface
from wiliot_deployment_tools.interface.uart_ports import get_uart_ports

def main():
    parser = ArgumentParser(prog='wlt-sniffer',
                            description='BLE Sniffer - CLI Tool to Sniff BLE Packets')
    required = parser.add_argument_group('required arguments')
    required.add_argument('-p', type=str, required=True, help=f'UART Port. Available ports: {str(get_uart_ports())}')
    required.add_argument('-c', type=int, help="channel", required=True, choices=[37, 38, 39])
    optional = parser.add_argument_group('optional arguments')
    optional.add_argument('-adva', type=str, required=False, help='ADVA filter for sniffer')
    env_dirs = WiliotDir()
    current_datetime = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
    sniffer_dir = os.path.join(env_dirs.get_wiliot_root_app_dir(), 'ble-sniffer', current_datetime)
    env_dirs.create_dir(sniffer_dir)
    start_timestamp = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
    sniffer_logger_filepath = os.path.join(sniffer_dir, f'{start_timestamp}_sniffer.log')

    
    args = parser.parse_args()
    uart = UARTInterface(args.p, update_fw=True)
    sniffer = BLESniffer(uart, print_pkt=True, logger_filepath=sniffer_logger_filepath, adva_filter=args.adva)
    sniffer.start_sniffer(args.c)
    while True:
        try:
            time.sleep(0.1)
        except KeyboardInterrupt:
            sniffer.stop_sniffer()
            break
        
def main_cli():
    main()


if __name__ == '__main__':
    main()



from wiliot_deployment_tools.internal.kafka_consumer.kafka_consumer import WiliotKafkaClient
from wiliot_deployment_tools.api.extended_api import BridgeThroughGatewayAction, ExtendedEdgeClient, GatewayType,ExtendedPlatformClient,EXCEPTIONS_TO_CATCH,BoardTypes
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.common.utils_defines import  SEC_TO_SEND,BROADCAST_DST_MAC
import json


class ConnectivtyAnalyzer():
    def __init__(self, edge_api_key, owner ,brgs = None, gws = None, env='prod', cloud="", region="us-east-2"):
        
        self.edge = ExtendedEdgeClient(edge_api_key, owner, env, cloud=cloud, region=region)
        self.gws = gws
        self.brgs = brgs
        self.kc = WiliotKafkaClient(env, cloud, keys=gws)
    
    def is_hexadecimal(self, char):
        """
        Check if a character is a hexadecimal digit.
        """
        return char.isdigit() or ('a' <= char.lower() <= 'f')
        
        
    def process_payload(self, input_string):
        """
        Process the string and extract the first 12 hexadecimal characters.
        """
        result = ''
        count = 0
        for char in input_string:
            if self.is_hexadecimal(char):
                result += char
                count += 1
                if count == 12:
                    break
        return result
    
    def send_hb_action(self, gateways):
        '''
        :param gateways: list of gateways to send action from
        Sends HB action from gateways to all seen bridges
        '''
        gw_payloads = []
        bridgeId = BROADCAST_DST_MAC
        for gw in gateways:
            try:
                gw_type = self.edge.get_gateway_type(gw)
                if gw_type is not GatewayType.ANDROID and gw_type is not GatewayType.MOBILE:
                    payload = self.process_payload(gw)
                    gw_payloads.append(payload)
                    try:
                        res = self.edge.send_bridge_action_through_gw(gw, BridgeThroughGatewayAction.GW_HB, payload, bridgeId)
                        debug_print(f'Sending HB packet through WIFI GW {gw}')
                    except Exception as e:
                        continue
                else:
                    debug_print(f'Gateway {gw} is a mobile gateway - not supported!')
            except Exception as e:
                if "not owned" in str(e):
                        debug_print(f"Warning!! gateway {gw} is not owned")
        return gw_payloads
    
    def check_ack(self, gw_payloads, bridges, gws):
        '''
        :param gw_payloads: A list of the payloads sent in the HB action by the gateway.
        :param bridges: A list of all bridges seen by the gateways.
        :return: A list of dictionaries with gw_payload, bridgeId and rssi as recieved in the gateway from
        that brideId's ack to the HB action.
        '''
        num_msgs_to_retreive = 500
        data = []
        messages = self.kc.get_last_messages_by_key(gws, num_msgs_to_retreive)
        for message in messages:
            value_str = message['value']
            try:
                value_dict = json.loads(value_str)
            except json.JSONDecodeError as e:
                debug_print(f"Error decoding value string: {e}")
                continue  # Skip to the next message if value_str is not in JSON format
            for key, value in value_dict.items():
                if key == 'packets':
                    parsed_data = {}
                    for packet in value:
                        remove_prefix_1E16 = lambda variable: str(variable)[4:] if str(variable).startswith("1E16") else variable
                        packet['payload'] = remove_prefix_1E16(packet['payload'])
                        payload_brg = packet['payload'][16:28]
                        payload_action_id = packet['payload'][28:30]
                        payload_gw = packet['payload'][30:42]
                        payload_rssi = int(packet['payload'][-2:], 16)
                        if payload_action_id == BridgeThroughGatewayAction.GW_HB.value and payload_gw in gw_payloads and payload_brg in bridges:
                            parsed_data = {'gw_payload' : payload_gw, 'bridge' : payload_brg, 'rssi' : payload_rssi}
                            data.append(parsed_data)   
        return data 
    
    def get_brgs_connected_gws(self, brgs):
        gws = []
        try:
            gws_dic = self.edge.get_gateways_from_bridges(brgs)
            for key, value in gws_dic.items():
                gws.append(key)
        except Exception as e:
            debug_print(f'Bridges not owned!; exception raised: {e}')
        return gws
    
    def get_gws_connected_brgs(self, gws):
        bridges = []
        for gw in gws:
            try:
                brgs_dic = self.edge.get_bridges(True, gw)
                for dic in brgs_dic:
                    bridges.append(dic['id'])
            except Exception as e:
                debug_print(f'Gateway {gw} is not owned!')
                continue
        return bridges
    
    def create_dic(self, gws, brgs):
        gw_brg_payload_dicts = []
        for gw in gws:
            brg_check_gw_list = []
            try:            
                gw_type = self.edge.get_gateway_type(gw)
                brg_check_gw_list.append(gw)
                brg_check = self.get_gws_connected_brgs(brg_check_gw_list)
                if gw_type is not GatewayType.ANDROID and gw_type is not GatewayType.MOBILE:
                    payload = self.process_payload(gw)
                else:
                    debug_print(f'Gateway {gw} is a mobile gateway - not supported!')
                for brg in brgs:
                    if brg in brg_check:
                        gw_brg_payload_dict = {'gw' : gw, 'bridge' : brg, 'gw_payload' : payload, 'rssi' : []}
                        gw_brg_payload_dicts.append(gw_brg_payload_dict)                    
            except Exception as e:
                if "not owned" in str(e):
                    debug_print(f"Warning!! gateway {gw} is not owned")
                    continue
        return gw_brg_payload_dicts
        
    def connectivity_analyzer(self):
        if self.gws and self.brgs:
            debug_print('Enter only brgs or gws')
            return None
        
        if self.gws:
            gws = self.gws
            gws_online = self.edge.check_gw_online(gws)
            if gws_online:
                brgs = self.get_gws_connected_brgs(gws)
                if brgs:
                    gw_brg_payload_dicts = self.create_dic(gws, brgs)
                else:
                    debug_print('No bridges connected to gateways for testing!')
            else:
                debug_print('Gateways offline!')
    
        elif self.brgs:
            brgs = self.brgs
            for brg in brgs:
                brg_online = self.edge.get_bridge_status(brg)
                if brg_online == 'offline':
                    brgs.remove(brg)
                    debug_print(f'Bridge {brg} is offline!')
            gws = self.get_brgs_connected_gws(brgs)
            if brgs and gws:
                gw_brg_payload_dicts = self.create_dic(gws, brgs)
            else:
                debug_print('No compatible gateways connected for running the test!')
                
        if gw_brg_payload_dicts:    
            gw_payloads = self.send_hb_action(gws)
            data = self.check_ack(gw_payloads, brgs, gws)
            brgs_with_good_uplink = []
            brgs_with_bad_uplink = []
            
            for dic_keys in gw_brg_payload_dicts:
                for dic_data in data:
                    if dic_data['gw_payload'] == dic_keys['gw_payload'] and dic_data['bridge'] == dic_keys['bridge']:
                        dic_keys['rssi'].append(dic_data['rssi'])
                if dic_keys['rssi'] and min(dic_keys['rssi']) < 76:
                    debug_print(f"Bridge {dic_keys['bridge']} has good connection to GW {dic_keys['gw']} with RSSIs {dic_keys['rssi']}")
                    brgs_with_good_uplink.append(dic_keys['bridge'])
                else:
                    debug_print(f"Bridge {dic_keys['bridge']} doesn't have a good connection to GW {dic_keys['gw']} with rssi {dic_keys['rssi']}")
                    brgs_with_bad_uplink.append(dic_keys['bridge'])
                    
            if brgs_with_good_uplink:
                for brg in list(set(brgs_with_good_uplink)):
                    debug_print(f'Bridge {brg} has a good downlink')
            if brgs_with_bad_uplink:
                for brg in list(set(brgs_with_bad_uplink)):
                    if brg not in brgs_with_good_uplink:
                        debug_print(f'Bridge {brg} has a bad downlink')
            return gw_brg_payload_dicts


from argparse import ArgumentParser
import time
from wiliot_deployment_tools.interface.ble_simulator import BLESimulator
from wiliot_deployment_tools.interface.if_defines import DEFAULT_DELAY, DEFAULT_DUPLICATES, DEFAULT_OUTPUT_POWER, SEND_ALL_ADV_CHANNELS
from wiliot_deployment_tools.interface.uart_ports import get_uart_ports
from wiliot_deployment_tools.interface.uart_if import UARTInterface

def main():
    parser = ArgumentParser(prog='wlt-ble-sim',
                            description='BLE Simulator - CLI Tool to use GW BLE Simulator')
    required = parser.add_argument_group('required arguments')
    required.add_argument('-p', type=str, help=f"USB UART Port (Available ports: {str(get_uart_ports())})", required=True)
    required.add_argument('-packet', type=str, help="packet", required=True)
    
    optional = parser.add_argument_group('optional arguments')
    optional.add_argument('-c', type=int, help="channel (if not specified packet will be sent on all BLE adv. channels (37/38/39))", default=SEND_ALL_ADV_CHANNELS)
    optional.add_argument('-delay', type=int, help="ms delay between packets (if not specified defaults to 20ms)", default=DEFAULT_DELAY)
    optional.add_argument('-dup', type=int, help="duplicates (defaults to 3)", default=DEFAULT_DUPLICATES)
    optional.add_argument('-output', type=int, help="output power (defaults to 8dBm)", default=DEFAULT_OUTPUT_POWER)
    optional.add_argument('-ts', type=float, help="trigger by time stamp") 
    
    args = parser.parse_args()
    u = UARTInterface(args.p, update_fw=True)
    b = BLESimulator(u)
    b.set_sim_mode(True)
    b.trigger_by_time_stamp(args.ts)
    b.send_packet(args.packet, args.dup, args.output, args.c, args.delay)
    b.set_sim_mode(False)
    
def main_cli():
    main()


if __name__ == '__main__':
    main()



from wiliot_deployment_tools.gw_certificate.tests.downlink import DownlinkTest 
from wiliot_deployment_tools.gw_certificate.tests.connection import ConnectionTest
from wiliot_deployment_tools.gw_certificate.tests.uplink import UplinkTest
from wiliot_deployment_tools.gw_certificate.tests.throughput import StressTest

TESTS = [ConnectionTest, UplinkTest, DownlinkTest, StressTest]





import datetime
import json
import time
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.gw_certificate.api_if.gw_capabilities import GWCapabilities
from wiliot_deployment_tools.gw_certificate.tests.generic import INCONCLUSIVE_MINIMUM, PassCriteria, MINIMUM_SCORE, PERFECT_SCORE, GenericStage, GenericTest, INFORMATIVE
from wiliot_deployment_tools.gw_certificate.api_if.api_validation import validate_message, MESSAGE_TYPES
from wiliot_deployment_tools.interface.mqtt import MqttClient, Serialization
from wiliot_deployment_tools.interface.ble_sniffer import BLESniffer, BLESnifferContext
import pkg_resources
import pandas as pd
from packaging import version
from wiliot_deployment_tools.gw_certificate.tests.static.connection_defines import *


class ConnectionStage(GenericStage):
    def __init__(self, mqttc:MqttClient, **kwargs):
        self.mqttc = mqttc
        self.stage_tooltip = "Awaits the gateway to establish MQTT connection and upload it's configurations via the 'status' topic as it's first message"
        self.__dict__.update(kwargs)
        super().__init__(stage_name=type(self).__name__, **self.__dict__)

    def run(self):
        self.stage_pass = MINIMUM_SCORE
        input('The GW is expected to publish a configuration JSON/Protobuf message through the status topic upon connecting to mqtt:\n'
                'Please unplug GW from power. Press enter when unplugged')
        self.mqttc.flush_messages()
        input('Please plug GW back to power. Press enter when plugged')
        debug_print('Waiting for GW to connect... (Timeout 3 minutes)')
        timeout = datetime.datetime.now() + datetime.timedelta(minutes=3)
        status_message = None
        while datetime.datetime.now() < timeout and status_message is None:
            status_message = self.mqttc.get_status_message()
        if status_message is not None:
            ser = self.mqttc.get_serialization()
            debug_print(f'-{ser.value} Serialization')
            self.add_to_stage_report(f'-{ser.value} Serialization')
            self.add_to_stage_report('GW Status packet received:')
            debug_print(status_message)
            self.add_to_stage_report(json.dumps(status_message))
            if ser == Serialization.JSON:
                validation = validate_message(MESSAGE_TYPES.STATUS, status_message)
                self.stage_pass = PERFECT_SCORE if validation[0] else MINIMUM_SCORE
            else:
                self.stage_pass = PERFECT_SCORE
            # set GW Capabilities:
            for key, value in status_message.items():
                if key in GWCapabilities.get_capabilities() and type(value) is bool:
                    self.gw_capabilities.set_capability(key, value)
                    self.add_to_stage_report(f'Set Capability: {key} - {value}')
            # Add reason test failed to report if neccessary
            if self.stage_pass == MINIMUM_SCORE:
                self.error_summary = "API (JSON structure) is invalid. "
                self.add_to_stage_report(f'{len(validation[1])} validation errors:')
                for error in validation[1]:
                    self.add_to_stage_report(error.message)
        else:
            error_summary = "No message recieved from GW in status topic after 3 mins."
            self.add_to_stage_report(error_summary)
            self.error_summary = error_summary

    def generate_stage_report(self):
        self.report_html = self.template_engine.render_template('stage.html', stage=self,
                                                                stage_report=self.report.split('\n'))
        debug_print(self.report)
        return super().generate_stage_report()
    

class InterferenceAnalysisStage(GenericStage):
    def __init__(self, sniffer:BLESniffer, **kwargs):
        self.sniffer = sniffer
        self.conversion_table_df = None
        self.stage_tooltip = "Analyze BLE interference level (Bad CRC %)"
        self.__dict__.update(kwargs)

        # Stage shows warning if CER is >=50%
        self.result_indication = INFORMATIVE
        self.pass_min = 51

        super().__init__(stage_name=type(self).__name__, **self.__dict__)

    def get_data_from_quantization_csv(self):
        relative_path = CSV_NAME
        csv_path = pkg_resources.resource_filename(__name__, relative_path)
        conversion_table_df = pd.read_csv(csv_path)
        self.conversion_table_df = conversion_table_df

    def interference_analysis(self):
        """Analyze the interference level (PER) before the test begins"""
        
        def handle_wrap_around(a):
            "handle a wrap around of the counter"
            if a < 0:
                a = a + MAX_UNSIGNED_32_BIT
            return a
        
        for channel in CHANNELS_TO_ANALYZE:
            # Send the sniffer a command to retrive the counters and convert them to dict
            start_cntrs = self.sniffer.get_pkts_cntrs(channel[0])
            debug_print(f'Analyzing channel {channel[0]}... (30 seconds)')
            time.sleep(CNTRS_LISTEN_TIME_SEC)
            end_cntrs = self.sniffer.get_pkts_cntrs(channel[0])

            if start_cntrs == None or end_cntrs == None:
                debug_print(f'Channel {channel[0]} ({channel[1]} MHz) interference analysis was skipped beacause at least one counter is missing.')
                self.add_to_stage_report(f'Channel {channel[0]} ({channel[1]} MHz) Ambient Interference was not calculated, missing at least one counter.')
                self.stage_pass = INCONCLUSIVE_MINIMUM
                continue

            # Calculate the bad CRC percentage
            diff_dict = dict()
            for key in CNTRS_KEYS:
                diff_dict[key] = handle_wrap_around(end_cntrs[key] - start_cntrs[key])
            if (diff_dict[WLT_RX] + diff_dict[NON_WLT_RX]) > 0:
                bad_crc_percentage = round((diff_dict[BAD_CRC] / (diff_dict[WLT_RX] + diff_dict[NON_WLT_RX])) * 100)
            else:
                bad_crc_percentage = 0
            self.add_to_stage_report(f'Channel {channel[0]} ({channel[1]} MHz) Ambient Interference (bad CRC percentage) is: {bad_crc_percentage}%.')
            self.add_to_stage_report(f'Good CRC packets = {diff_dict[NON_WLT_RX] + diff_dict[WLT_RX] - diff_dict[BAD_CRC]}, bad CRC packets: {diff_dict[BAD_CRC]}')

            good_crc_percentage = 100 - bad_crc_percentage
            if (self.stage_pass == MINIMUM_SCORE) or (good_crc_percentage < self.stage_pass):
                self.stage_pass = good_crc_percentage
                if self.stage_pass < self.pass_min:
                    self.error_summary = "High bad CRC rate within the current environment."
        
            # Uncomment if you want to see PER of the site (will require print adjustments). Below, we use the truth table from the csv to match PER the bad CRC percentage. Require an update of the CSV to the bridge-GW case
            # closest_index = (self.conversion_table_df['bad_crc_percent'] - bad_crc_percentage).abs().idxmin()
            # per_percent = self.conversion_table_df.iloc[closest_index]['per_percent']
            # self.add_to_stage_report(f'Channel {channel} PER is: {per_percent}%')

    def run(self):
        # Run interference analysis
        # Note: there is an infrastructure for converting bad_CRC % to PER, currently unused and commented since the quantization_csv does not match the bridge to GW case.
        debug_print(f"Starting interference analysis for channels {[ch[0] for ch in CHANNELS_TO_ANALYZE]}. This will take {30 * len(CHANNELS_TO_ANALYZE)} seconds (total)")
        # self.get_data_from_quantization_csv()
        self.interference_analysis()
    
    def generate_stage_report(self):
        self.report_html = self.template_engine.render_template('stage.html', stage=self,
                                                                stage_report=self.report.split('\n'))
        debug_print(self.report)
        return super().generate_stage_report()

STAGES = [ConnectionStage]

class ConnectionTest(GenericTest):
    def __init__(self, **kwargs):
        self.test_tooltip = "Stages related to cloud connectivity and environment"
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, test_name=type(self).__name__)
        stages = STAGES
        if version.parse(self.uart.get_version()) >= version.parse(INTERFERENCE_ANALYSIS_FW_VER):
            stages.append(InterferenceAnalysisStage)
        self.stages = [stage(**self.__dict__) for stage in stages]
        
    def run(self):
        super().run()
        self.test_pass = PERFECT_SCORE
        for stage in self.stages:
            stage.prepare_stage()
            stage.run()
            self.test_pass = PassCriteria.calc_for_test(self.test_pass, stage)
            self.add_to_test_report(stage.generate_stage_report())



import copy
import datetime
from enum import Enum
import json
import logging
import time
from typing import Literal, Union
import uuid
import paho.mqtt.client as mqtt
import ssl
import base64
from wiliot_deployment_tools.ag.ut_defines import PACKETS, PAYLOAD, MGMT_PKT, SIDE_INFO_PKT, GW_ID, NFPKT, GW_LOGS, UNIFIED_PKT
from wiliot_deployment_tools.ag.wlt_types_ag import GROUP_ID_BRG2GW, GROUP_ID_SIDE_INFO, GROUP_ID_UNIFIED_PKT
from wiliot_deployment_tools.ag.wlt_types_data import DATA_DEFAULT_GROUP_ID, DataPacket
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.common import wltPb_pb2
from google.protobuf.message import DecodeError
from google.protobuf.json_format import MessageToDict, Parse, ParseError, ParseDict



DATA_PKT = 'data_pkt'

class CustomBrokers(Enum):
    HIVE = 'broker.hivemq.com'
    EMQX = 'broker.emqx.io'
    ECLIPSE = 'mqtt.eclipseprojects.io'
    
def get_broker_url(broker):
        try:
            broker_url = CustomBrokers[broker.upper()].value
            debug_print(f"Broker URL: {broker_url}")
            return broker_url
        except KeyError:
            raise KeyError(f"Broker '{broker}' not found in CustomBrokers.")

class Serialization(Enum):
    UNKNOWN = "unknown"
    JSON = "JSON"
    PB = "Protobuf"

class GwAction(Enum):
    DISABLE_DEV_MODE = "DevModeDisable"
    REBOOT_GW ="rebootGw"
    GET_GW_INFO ="getGwInfo"

class WltMqttMessage:
    def __init__(self, body, topic):
        self.body = body
        self.mqtt_topic = topic
        self.mqtt_timestamp = datetime.datetime.now()
        self.body_ex = copy.deepcopy(body)
        self.is_unified = False
        if "data" in self.mqtt_topic and PACKETS in self.body_ex.keys():
            for pkt in self.body_ex[PACKETS]:
                data_pkt = DataPacket()
                data_pkt.set(pkt[PAYLOAD])
                if data_pkt.pkt != None:
                    if data_pkt.hdr.group_id == GROUP_ID_BRG2GW:
                        pkt[MGMT_PKT] = copy.deepcopy(data_pkt)
                if data_pkt.hdr.group_id == GROUP_ID_SIDE_INFO:
                    pkt[SIDE_INFO_PKT] = copy.deepcopy(data_pkt)
                if data_pkt.hdr.group_id == DATA_DEFAULT_GROUP_ID:
                    pkt[DATA_PKT] = copy.deepcopy(data_pkt)
                if data_pkt.hdr.group_id == GROUP_ID_UNIFIED_PKT:
                    pkt[UNIFIED_PKT] = copy.deepcopy(data_pkt)
                    self.is_unified = True

    def __repr__(self) -> str:
        if self.body_ex != {}:
            return str(self.body_ex)
        return str(self.body)


class WltMqttMessages:
    def __init__(self):
        self.data = []
        self.status = []
        self.update = []
        self.all = []

    def insert(self, pkt):
        self.all.append(pkt)
        if "data" in pkt.mqtt_topic:
            self.data.append(pkt)
        elif "status" in pkt.mqtt_topic:
            self.status.append(pkt)
        elif "update" in pkt.mqtt_topic:
            self.update.append(pkt)
            
    def __repr__(self) -> str:
        return f'Data {self.data} \n Status {self.status} \n Update {self.update}'
            
class MqttClient:

    def __init__(self, gw_id, owner_id, logger_filepath=None, topic_suffix='', serialization=Serialization.UNKNOWN, broker='hive'):
        # Set variables
        self.gw_id = gw_id
        self.owner_id = owner_id
        self.broker_url = get_broker_url(broker)
        
        # Configure logger
        logger = logging.getLogger('mqtt')
        logger.setLevel(logging.DEBUG)
        if logger_filepath is not None:
            # create file handler which logs even debug messages
            fh = logging.FileHandler(logger_filepath)
            fh.setLevel(logging.DEBUG)
            formatter = logging.Formatter('%(asctime)s | %(message)s')
            fh.setFormatter(formatter)
            logger.addHandler(fh)
            logger.propagate = False # Do not send logs to 'root' logger
            debug_print(f'MQTT Logger initialized at {logger_filepath}')
        self.logger = logger
        
        # Configure Paho MQTT Client
        client_id = f'GW_Certificate_{uuid.uuid4()}'
        self.userdata = {'messages': WltMqttMessages(), 'gw_seen': False , 'logger': self.logger, 'serialization': serialization, 'published': []}
        # Try-except is temporary until old users are up to date
        try:
            self.client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2, client_id, userdata=self.userdata)
        except AttributeError:
            print("\nGW Certificate now runs with latest paho-mqtt!\nPlease upgrade yours to version 2.0.0 (pip install --upgrade paho-mqtt)\n")
            raise
        self.client.enable_logger(logger=self.logger)
        self.client.on_message = on_message
        self.client.on_connect = on_connect
        self.client.on_disconnect = on_disconnect
        self.client.on_subscribe = on_subscribe
        self.client.on_unsubscribe = on_unsubscribe
        self.client.on_publish = on_publish
        self.client.on_log = on_log
        self.client.tls_set(tls_version=ssl.PROTOCOL_TLSv1_2)
        debug_print(f'Connecting to MQTT broker: tls://{self.broker_url}:8883, Keepalive=60')
        self.client.connect(self.broker_url, port=8883, keepalive=60)
        # Set Topics
        self.update_topic = f"update{topic_suffix}/{owner_id}/{gw_id}"
        debug_print(f'Subscribe to {self.update_topic}...')
        self.client.subscribe(self.update_topic)
        self.data_topic = f"data{topic_suffix}/{owner_id}/{gw_id}"
        debug_print(f'Subscribe to {self.data_topic}...')
        self.client.subscribe(self.data_topic)
        self.status_topic = f"status{topic_suffix}/{owner_id}/{gw_id}"
        debug_print(f'Subscribe to {self.status_topic}...')
        self.client.subscribe(self.status_topic)
        self.client.loop_start()
        while(not self.client.is_connected()):
            debug_print(f'Waiting for MQTT connection...')
            time.sleep(1)
        debug_print('Connected to MQTT.')
    
    def get_serialization(self):
        """
        return serialization type
        """
        return self.userdata['serialization']
    
    # Downstream Interface
    def send_action(self, action:GwAction):
        """
        Send an action to the gateway
        :param action: GwAction - Required
        """
        assert isinstance(action, GwAction), 'Action Must be a GWAction!'
        # JSON
        if self.get_serialization() in {Serialization.UNKNOWN, Serialization.JSON}:
            raw_payload = json.dumps({"action": action.value})
            self.userdata['published'].append(raw_payload.encode('utf-8'))
            message_info = self.client.publish(self.update_topic, payload=raw_payload)
            print('PublishJSON')
        # PB
        if self.get_serialization() in {Serialization.UNKNOWN, Serialization.PB}:
            payload = wltPb_pb2.DownlinkMessage()
            payload.gatewayAction.action = action.value
            raw_payload = payload.SerializeToString()
            self.userdata['published'].append(raw_payload)
            message_info = self.client.publish(self.update_topic, payload=raw_payload)
            print('PublishPB')
        message_info.wait_for_publish()
        return message_info
    
    def send_payload(self, payload, topic:Literal['update', 'data', 'status']='update'):
        """
        Send a payload to the gateway
        :type payload: dict [JSON] / str [PB]
        :param payload: payload to send
        :type topic: Literal['update', 'data', 'status']
        :param topic: defualts to update
        """
        def cast_to_proto(payload: Union[str, dict, wltPb_pb2.DownlinkMessage]):
            if isinstance(payload, wltPb_pb2.DownlinkMessage):
                return payload

            if isinstance(payload, str):
                payload = json.loads(payload)
            # payload is now a dictionary

            def add_proto_field_type(d):
                for key, value in list(d.items()):
                    if isinstance(value, int):
                        d[key] = {"integerValue": value}
                    elif isinstance(value, float):
                        d[key] = {"numberValue": value}
                    elif isinstance(value, str):
                        d[key] = {"stringValue": value}
                    elif isinstance(value, bool):
                        d[key] = {"boolValue": value}
                    elif isinstance(value, dict):
                        add_proto_field_type(value)  # Recursively handle nested dictionaries

            if 'gatewayConfig' in payload:
                config = payload['gatewayConfig'].get('config', {})
                add_proto_field_type(config)
            
            pb_message = wltPb_pb2.DownlinkMessage()
            ParseDict(payload, pb_message, ignore_unknown_fields=True)

            return pb_message
            
        topic = {'update': self.update_topic,
                 'data': self.data_topic, 
                 'status': self.status_topic}[topic]
        # JSON
        if self.get_serialization() in {Serialization.UNKNOWN, Serialization.JSON}:
            try:
                raw_payload = json.dumps(payload)
                # Add published payload to published list
                self.userdata['published'].append(raw_payload.encode('utf-8'))
                message_info = self.client.publish(topic, raw_payload)
            except TypeError as e:
                if self.get_serialization() != Serialization.UNKNOWN:
                    debug_print(f'Cannot pack payload as JSON!: {payload}')
                    raise e
        # PB
        if self.get_serialization() in {Serialization.UNKNOWN, Serialization.PB}:
            try:
                pb_message = cast_to_proto(payload)
                raw_payload = pb_message.SerializeToString()
                # Add published payload to published list
                self.userdata['published'].append(raw_payload)
                message_info = self.client.publish(topic, raw_payload)
            except ParseError as e:
                if self.get_serialization() != Serialization.UNKNOWN:
                    debug_print(f'Cannot parse payload as PB message!: {payload}')
                    raise e
        message_info.wait_for_publish()
        return message_info
    
    def advertise_packet(self, raw_packet, tx_max_duration=800, use_retries=False):
        if len(raw_packet) < 62:
            if len(raw_packet) == 54:
                raw_packet = 'C6FC' + raw_packet
            if len(raw_packet) == 58:
                raw_packet = '1E16' + raw_packet
        if len(raw_packet) > 62:
            raw_packet = raw_packet[-62:]
        
        assert len(raw_packet) == 62, 'Raw Packet must be 62 chars long!'

        if self.get_serialization() == Serialization.PB: # PB Serialization
            payload = wltPb_pb2.DownlinkMessage()
            payload.txPacket.payload = bytes.fromhex(raw_packet)
            payload.txPacket.maxRetries = int(tx_max_duration / 100)
            payload.txPacket.maxDurationMs = tx_max_duration
        
        else: # JSON Serialization    
            if use_retries:
                payload = {
                    'action': 0, # Advertise BLE Packet
                    'txPacket': raw_packet, # Raw Packet
                    'txMaxRetries': tx_max_duration / 100, # Tx Max Retries
                }
            else:
                payload = {
                'txPacket': raw_packet, # Raw Packet
                'txMaxDurationMs': tx_max_duration, # Tx Max Duration
                'action': 0 # Advertise BLE Packet
            }
        
        self.send_payload(payload, topic='update')
        return payload

    def check_gw_seen(self):
        return self.userdata['gw_seen']
    
    def get_gw_info(self):
        self.flush_messages()
        self.send_action(GwAction.GET_GW_INFO)
        time.sleep(5)
        debug_print('---GW INFO---')
        try:
            gw_info = self.userdata['messages'].status[0]
            debug_print(gw_info)
            self.flush_messages()
            return gw_info
        except IndexError:
            debug_print('No GW INFO')
            self.flush_messages()
            return False
        
    def get_gw_configuration(self):
        self.flush_messages()
        self.send_action(GwAction.REBOOT_GW)
        debug_print('---GW CONFIG---')
        try:
            debug_print(self.userdata['messages'].status)
            return True
        except KeyError:
            return False
            
    def exit_custom_mqtt(self, mqtt_mode:Literal['automatic', 'manual' ,'legacy']):
        if mqtt_mode == 'legacy':
            return self.send_action(GwAction.DISABLE_DEV_MODE)
        elif mqtt_mode == 'automatic':
            custom_mqtt = {
                "customBroker": False,
                "brokerUrl": "",
                "port": 8883,
                "username": "",
                "password": "",
                "updateTopic": f"update/{self.owner_id}/{self.gw_id}",
                "statusTopic": f"status/{self.owner_id}/{self.gw_id}",
                "dataTopic": f"data/{self.owner_id}/{self.gw_id}"
                }
            if self.get_serialization() in {Serialization.UNKNOWN, Serialization.JSON}:
                self.send_payload(custom_mqtt)
            if self.get_serialization() in {Serialization.UNKNOWN, Serialization.PB}:
                self.send_payload({'customBroker': custom_mqtt})
        elif mqtt_mode == 'manual':
            debug_print(f"Make sure GW {self.gw_id} is set to Wiliot MQTT broker")
            return True
    
    # Packet Handling
    def flush_messages(self):
        self.userdata = {'messages': WltMqttMessages(), 'gw_seen': False , 'logger': self.logger, 'serialization':self.get_serialization(), 'published':[]}
        self.client.user_data_set(self.userdata)
    
    def get_all_messages_from_topic(self, topic:Literal['status', 'data', 'update']):
        return getattr(self.userdata['messages'], topic)
    
    def get_all_pkts_from_topic(self, topic:Literal['status', 'data', 'update']):
        pkts = []
        if topic == 'data':
            for p in eval(f'self.userdata["messages"].{topic}'):
                gw_id = p.body_ex[GW_ID] if GW_ID in p.body_ex else ""
                if PACKETS in p.body_ex:
                    for pkt in p.body_ex[PACKETS]:
                        pkt[GW_ID] = gw_id
                        pkts += [pkt]
            return pkts
    
    def get_status_message(self):
        messages = self.get_all_messages_from_topic('status')
        for message in messages:
            if GW_LOGS not in message.body_ex.keys():
                return message.body_ex
        return None
    
    def get_coupled_tags_pkts(self):
        return [p for p in self.get_all_pkts_from_topic('data') if NFPKT in p]
    
    def get_uncoupled_tags_pkts(self):
        return [p for p in self.get_all_pkts_from_topic('data') if NFPKT not in p]
    
    def get_all_tags_pkts(self):
        return [p for p in self.get_all_pkts_from_topic('data')]

# MQTT Client callbacks

def on_connect(mqttc, userdata, flags, reason_code, properties):
    message = f'MQTT: Connection, RC {reason_code}'
    userdata['logger'].info(message)
    # Properties and Flags
    userdata['logger'].info(flags)
    userdata['logger'].info(properties)

def on_disconnect(mqttc, userdata, flags, reason_code, properties):
    if reason_code != 0:
        userdata['logger'].info(f"MQTT: Unexpected disconnection. {reason_code}")
    else:
        userdata['logger'].info('MQTT: Disconnect')
    userdata['logger'].info(flags)
    userdata['logger'].info(properties)

def on_subscribe(mqttc, userdata, mid, reason_codes, properties):
    userdata['logger'].info(f"MQTT: Subscribe, MessageID {mid}")
    for sub_result, idx in enumerate(reason_codes):
        userdata['logger'].info(f"[{idx}]: RC {sub_result}")
    userdata['logger'].info(properties)

def on_unsubscribe(mqttc, userdata, mid, reason_codes, properties):
    userdata['logger'].info(f"MQTT: Unsubscribe, MessageID {mid}")
    for sub_result, idx in enumerate(reason_codes):
        userdata['logger'].info(f"[{idx}]: RC {sub_result}")
    userdata['logger'].info(properties)

def on_message(mqttc, userdata, message):
    # Ignore messages published by MQTT Client
    if message.payload in userdata['published']:
        userdata['logger'].info(f'Received self-published payload - {message.topic}: {message.payload}')
        return
    # Try to parse message as JSON and determine if GW is working in JSON / PB mode
    if userdata['serialization'] == Serialization.UNKNOWN:
        try:
            payload = message.payload.decode("utf-8")
            userdata['logger'].info("Received JSON-Serialized packet - setting serialization to JSON")
            debug_print("##### Received JSON-Serialized packet - setting serialization to JSON #####")
            userdata['serialization'] = Serialization.JSON
        except (json.JSONDecodeError, UnicodeDecodeError):
            userdata['logger'].info("Received non-JSON-Serialized packet - setting serialization to PB")
            debug_print("##### Received non-JSON-Serialized packet - setting serialization to PB #####")
            userdata['serialization'] = Serialization.PB
    if userdata['serialization'] == Serialization.JSON:
        on_message_json(mqttc, userdata, message)
    if userdata['serialization'] == Serialization.PB:
        on_message_protobuf(mqttc, userdata, message)

def on_message_json(mqttc, userdata, message):
    payload = message.payload.decode("utf-8")
    data = json.loads(payload)
    userdata['messages'].insert(WltMqttMessage(data, message.topic))
    userdata['logger'].debug(f'{message.topic}: {payload}')
    if(userdata['gw_seen'] is False):
        userdata['gw_seen'] = True

def on_message_protobuf(mqttc, userdata, message):
    pb_message = None
    if 'status' in message.topic:
        # Try to decode UplinkMessage
        try:
            pb_message = wltPb_pb2.UplinkMessage()
            pb_message.ParseFromString(message.payload)
        except DecodeError as e: 
            userdata['logger'].debug(f'{message.topic}: An exception occured:\n{e}\n(could be a JSON msg or pb msg that is not UplinkMessage)###########')
            userdata['logger'].debug(f'Raw Payload: {message.payload}')
    elif 'data' in message.topic:
        # Try to decode GatewayData
        try:
            pb_message = wltPb_pb2.GatewayData()
            pb_message.ParseFromString(message.payload)
            # # Decode packets payloads as hex
            # for idx, packet in enumerate(pb_message.packets):
            #     pb_message.packets[idx].payload = packet.payload.hex().upper().encode('utf-8')
        except DecodeError as e: 
            userdata['logger'].debug(f'{message.topic}: An exception occured:\n{e}\n(could be a JSON msg or pb msg that is not UplinkMessage)###########')
            userdata['logger'].debug(f'Raw Payload: {message.payload}')
    else:
        userdata['logger'].debug(f'Message from update topic, not decoding')
        userdata['logger'].debug(f'{message.topic}: {message.payload}')
    if pb_message is not None:
        pb_message_dict = MessageToDict(pb_message)
        # Decode b64 packet payloads to hex
        if 'data' in message.topic and 'packets' in pb_message_dict.keys():
            for idx, packet in enumerate(pb_message_dict['packets']):
                pb_message_dict['packets'][idx]['payload'] = base64.b64decode(packet['payload']).hex().upper()
        userdata['messages'].insert(WltMqttMessage(pb_message_dict, message.topic))
        userdata['logger'].debug(f'{message.topic}: {pb_message.__class__.__name__}')
        userdata['logger'].debug(f'{pb_message_dict}')
        if(userdata['gw_seen'] is False):
            userdata['gw_seen'] = True

def on_publish(mqttc, userdata, mid, reason_code, properties):
    userdata['logger'].info(f"MQTT: Publish, MessageID {mid}, RC {reason_code}")
    userdata['logger'].info(properties)

def on_log(mqttc, userdata, level, buf):
    if (level < mqtt.MQTT_LOG_DEBUG):
        userdata['logger'].info(f"MQTT: Log level={level}, Msg={buf}")


import os
import json
from jsonschema import Draft202012Validator
from wiliot_deployment_tools.ag.ut_defines import GW_CONF, GW_API_VERSION
from wiliot_deployment_tools.common.debug import debug_print
import pkg_resources

from enum import Enum

class MESSAGE_TYPES(Enum):
    STATUS = "status"
    DATA = "data"
    LOGS = "logs"

# TODO Can we use GW_API_VER_DEFAULT from ag.ut_defines instead after we auto-generate again?
LATEST_API = "203"
api_version = LATEST_API

def validate_message(message_type: MESSAGE_TYPES, message: dict) -> tuple[bool, str]:
    """
    Validate MQTT message
    :type message_type: MESSAGE_TYPES
    :param message_type: MQTT message type
    :type message: dict
    :param message: MQTT Message
    :return: tuple (bool, str)
    """
    global api_version

    # NOTE When Connection test is not ran (using the -tests flag), default api_version is the latest
    if message_type == MESSAGE_TYPES.STATUS:
        api_version = message[GW_CONF][GW_API_VERSION]
        debug_print(f'API version set as {api_version} according to the status message.')
    json_path = pkg_resources.resource_filename(__name__, f"{api_version}/{message_type.value}.json")
    with open(json_path) as f:
        relevant_schema = json.load(f)
    validator = Draft202012Validator(relevant_schema)
    valid = validator.is_valid(message)
    errors = [e for e in validator.iter_errors(message)]
    return (valid, errors)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: wltPb.proto
# Protobuf Python Version: 5.28.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    2,
    '',
    'wltPb.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0bwltPb.proto\x12\x05wltPb\"$\n\x08Location\x12\x0b\n\x03lat\x18\x01 \x01(\x01\x12\x0b\n\x03lng\x18\x02 \x01(\x01\"j\n\x05Value\x12\x16\n\x0cintegerValue\x18\x01 \x01(\x03H\x00\x12\x15\n\x0bnumberValue\x18\x02 \x01(\x01H\x00\x12\x15\n\x0bstringValue\x18\x03 \x01(\tH\x00\x12\x13\n\tboolValue\x18\x04 \x01(\x08H\x00\x42\x06\n\x04type\"\x89\x02\n\x0bGatewayData\x12\x11\n\tgatewayId\x18\x01 \x01(\t\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\x12*\n\x07packets\x18\x03 \x03(\x0b\x32\x19.wltPb.GatewayData.Packet\x12&\n\x08location\x18\x04 \x01(\x0b\x32\x0f.wltPb.LocationH\x00\x88\x01\x01\x1as\n\x06Packet\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\x12\x12\n\nsequenceId\x18\x03 \x01(\r\x12\x11\n\x04rssi\x18\x04 \x01(\rH\x00\x88\x01\x01\x12\x15\n\raliasBridgeId\x18\x05 \x01(\tB\x07\n\x05_rssiB\x0b\n\t_location\"\xc9\x01\n\rUplinkMessage\x12-\n\rgatewayStatus\x18\x01 \x01(\x0b\x32\x14.wltPb.GatewayStatusH\x00\x12)\n\x0bgatewayInfo\x18\x02 \x01(\x0b\x32\x12.wltPb.GatewayInfoH\x00\x12)\n\x0bgatewayLogs\x18\x03 \x01(\x0b\x32\x12.wltPb.GatewayLogsH\x00\x12+\n\x0c\x61\x63tionStatus\x18\x04 \x01(\x0b\x32\x13.wltPb.ActionStatusH\x00\x42\x06\n\x04type\"\xdb\x03\n\rGatewayStatus\x12\x11\n\tgatewayId\x18\x01 \x01(\t\x12\x13\n\x0bgatewayType\x18\x02 \x01(\t\x12\x19\n\x11\x64ownlinkSupported\x18\x03 \x01(\x08\x12!\n\x19\x62ridgeOtaUpgradeSupported\x18\x04 \x01(\x08\x12\x12\n\napiVersion\x18\x05 \x01(\r\x12\x14\n\x07version\x18\x06 \x01(\tH\x00\x88\x01\x01\x12\x19\n\x0c\x62leSwVersion\x18\x07 \x01(\tH\x01\x88\x01\x01\x12\x1f\n\x12interfaceSwVersion\x18\x08 \x01(\tH\x02\x88\x01\x01\x12&\n\x08location\x18\t \x01(\x0b\x32\x0f.wltPb.LocationH\x03\x88\x01\x01\x12\x30\n\x06\x63onfig\x18\n \x03(\x0b\x32 .wltPb.GatewayStatus.ConfigEntry\x12\x17\n\nbleAddress\x18\x0b \x01(\tH\x04\x88\x01\x01\x1a;\n\x0b\x43onfigEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1b\n\x05value\x18\x02 \x01(\x0b\x32\x0c.wltPb.Value:\x02\x38\x01\x42\n\n\x08_versionB\x0f\n\r_bleSwVersionB\x15\n\x13_interfaceSwVersionB\x0b\n\t_locationB\r\n\x0b_bleAddress\"}\n\x0bGatewayInfo\x12\x30\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x1f.wltPb.GatewayInfo.EntriesEntry\x1a<\n\x0c\x45ntriesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1b\n\x05value\x18\x02 \x01(\x0b\x32\x0c.wltPb.Value:\x02\x38\x01\"\x1b\n\x0bGatewayLogs\x12\x0c\n\x04logs\x18\x01 \x03(\t\".\n\x0c\x41\x63tionStatus\x12\x0e\n\x06\x61\x63tion\x18\x01 \x01(\r\x12\x0e\n\x06status\x18\x02 \x01(\r\"\xf8\x01\n\x0f\x44ownlinkMessage\x12#\n\x08txPacket\x18\x01 \x01(\x0b\x32\x0f.wltPb.TxPacketH\x00\x12-\n\rgatewayAction\x18\x02 \x01(\x0b\x32\x14.wltPb.GatewayActionH\x00\x12-\n\rbridgeUpgrade\x18\x03 \x01(\x0b\x32\x14.wltPb.BridgeUpgradeH\x00\x12-\n\rgatewayConfig\x18\x04 \x01(\x0b\x32\x14.wltPb.GatewayConfigH\x00\x12+\n\x0c\x63ustomBroker\x18\x05 \x01(\x0b\x32\x13.wltPb.CustomBrokerH\x00\x42\x06\n\x04type\"F\n\x08TxPacket\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\x12\x12\n\nmaxRetries\x18\x02 \x01(\r\x12\x15\n\rmaxDurationMs\x18\x03 \x01(\r\"\x1f\n\rGatewayAction\x12\x0e\n\x06\x61\x63tion\x18\x01 \x01(\t\"\xa5\x01\n\rBridgeUpgrade\x12\x14\n\x0crebootPacket\x18\x01 \x01(\x0c\x12\x17\n\x0ftxMaxDurationMs\x18\x02 \x01(\r\x12\x14\n\x0ctxMaxRetries\x18\x03 \x01(\r\x12\x10\n\x08\x62ridgeId\x18\x04 \x01(\t\x12\x13\n\x0bversionUuid\x18\x05 \x01(\t\x12\x13\n\x0bupgradeBlSd\x18\x06 \x01(\x08\x12\x13\n\x0bimageDirUrl\x18\x07 \x01(\t\"\xb9\x02\n\rGatewayConfig\x12\x14\n\x07version\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x19\n\x0c\x62leSwVersion\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x1f\n\x12interfaceSwVersion\x18\x03 \x01(\tH\x02\x88\x01\x01\x12&\n\x08location\x18\x04 \x01(\x0b\x32\x0f.wltPb.LocationH\x03\x88\x01\x01\x12\x30\n\x06\x63onfig\x18\x05 \x03(\x0b\x32 .wltPb.GatewayConfig.ConfigEntry\x1a;\n\x0b\x43onfigEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1b\n\x05value\x18\x02 \x01(\x0b\x32\x0c.wltPb.Value:\x02\x38\x01\x42\n\n\x08_versionB\x0f\n\r_bleSwVersionB\x15\n\x13_interfaceSwVersionB\x0b\n\t_location\"\xa6\x01\n\x0c\x43ustomBroker\x12\x14\n\x0c\x63ustomBroker\x18\x01 \x01(\x08\x12\x0c\n\x04port\x18\x02 \x01(\r\x12\x11\n\tbrokerUrl\x18\x03 \x01(\t\x12\x10\n\x08username\x18\x04 \x01(\t\x12\x10\n\x08password\x18\x05 \x01(\t\x12\x13\n\x0bupdateTopic\x18\x06 \x01(\t\x12\x13\n\x0bstatusTopic\x18\x07 \x01(\t\x12\x11\n\tdataTopic\x18\x08 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'wltPb_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GATEWAYSTATUS_CONFIGENTRY']._loaded_options = None
  _globals['_GATEWAYSTATUS_CONFIGENTRY']._serialized_options = b'8\001'
  _globals['_GATEWAYINFO_ENTRIESENTRY']._loaded_options = None
  _globals['_GATEWAYINFO_ENTRIESENTRY']._serialized_options = b'8\001'
  _globals['_GATEWAYCONFIG_CONFIGENTRY']._loaded_options = None
  _globals['_GATEWAYCONFIG_CONFIGENTRY']._serialized_options = b'8\001'
  _globals['_LOCATION']._serialized_start=22
  _globals['_LOCATION']._serialized_end=58
  _globals['_VALUE']._serialized_start=60
  _globals['_VALUE']._serialized_end=166
  _globals['_GATEWAYDATA']._serialized_start=169
  _globals['_GATEWAYDATA']._serialized_end=434
  _globals['_GATEWAYDATA_PACKET']._serialized_start=306
  _globals['_GATEWAYDATA_PACKET']._serialized_end=421
  _globals['_UPLINKMESSAGE']._serialized_start=437
  _globals['_UPLINKMESSAGE']._serialized_end=638
  _globals['_GATEWAYSTATUS']._serialized_start=641
  _globals['_GATEWAYSTATUS']._serialized_end=1116
  _globals['_GATEWAYSTATUS_CONFIGENTRY']._serialized_start=977
  _globals['_GATEWAYSTATUS_CONFIGENTRY']._serialized_end=1036
  _globals['_GATEWAYINFO']._serialized_start=1118
  _globals['_GATEWAYINFO']._serialized_end=1243
  _globals['_GATEWAYINFO_ENTRIESENTRY']._serialized_start=1183
  _globals['_GATEWAYINFO_ENTRIESENTRY']._serialized_end=1243
  _globals['_GATEWAYLOGS']._serialized_start=1245
  _globals['_GATEWAYLOGS']._serialized_end=1272
  _globals['_ACTIONSTATUS']._serialized_start=1274
  _globals['_ACTIONSTATUS']._serialized_end=1320
  _globals['_DOWNLINKMESSAGE']._serialized_start=1323
  _globals['_DOWNLINKMESSAGE']._serialized_end=1571
  _globals['_TXPACKET']._serialized_start=1573
  _globals['_TXPACKET']._serialized_end=1643
  _globals['_GATEWAYACTION']._serialized_start=1645
  _globals['_GATEWAYACTION']._serialized_end=1676
  _globals['_BRIDGEUPGRADE']._serialized_start=1679
  _globals['_BRIDGEUPGRADE']._serialized_end=1844
  _globals['_GATEWAYCONFIG']._serialized_start=1847
  _globals['_GATEWAYCONFIG']._serialized_end=2160
  _globals['_GATEWAYCONFIG_CONFIGENTRY']._serialized_start=977
  _globals['_GATEWAYCONFIG_CONFIGENTRY']._serialized_end=1036
  _globals['_CUSTOMBROKER']._serialized_start=2163
  _globals['_CUSTOMBROKER']._serialized_end=2329
# @@protoc_insertion_point(module_scope)



from argparse import ArgumentParser
from wiliot_deployment_tools.gw_certificate.gw_certificate import GWCertificate
from wiliot_deployment_tools.gw_certificate.tests import TESTS

def filter_tests(tests_names):
    chosen_tests = []
    if tests_names == []:
        return TESTS
    for test_class in TESTS:
        for test_name in tests_names:
            if test_name in test_class.__name__.lower() and test_class not in chosen_tests:
                chosen_tests.append(test_class)
    return chosen_tests

def main():
    usage = (
        "usage: wlt-gw-certificate [-h] -owner OWNER -gw GW\n"
        "                          [-tests {connection,uplink,downlink,stress}] [-update]"
        )

    parser = ArgumentParser(prog='wlt-gw-certificate',
                            description='Gateway Certificate - CLI Tool to test Wiliot GWs', usage=usage)

    required = parser.add_argument_group('required arguments')
    required.add_argument('-owner', type=str, help="Owner ID", required=True)
    required.add_argument('-gw', type=str, help="Gateway ID", required=True)
    optional = parser.add_argument_group('optional arguments')
    optional.add_argument('-suffix', type=str, help="Topic suffix", default='', required=False)
    optional.add_argument('-tests', type=str, choices=['connection', 'uplink', 'downlink', 'stress'], help="Tests to run", required=False, nargs='+', default=[])
    optional.add_argument('-update', action='store_true', help='Update test board firmware', default=False, required=False)
    args = parser.parse_args()
    tests = filter_tests(args.tests)
    topic_suffix = '' if args.suffix == '' else '-'+args.suffix


    gwc = GWCertificate(gw_id=args.gw, owner_id=args.owner, topic_suffix=topic_suffix, tests=tests, update_fw=args.update)
    gwc.run_tests()
    gwc.create_results_html()

def main_cli():
    main()

if __name__ == '__main__':
    main()
    


# Brownout / OTA Defines
SEP = "#" * 50
MINUTES_TO_BROWN_OUT = 5
TABLES_SYNC_MINUTES = 70
MINUTES_FOR_LC = 2
LATEST_FW = "1.9.0"
SINGLE_BAT_BO_DICT = {"2.4GhzOutputPower": 2, 'txPeriodMs': 75, 'rxTxPeriodMs': 255, 'energyPattern': 25}
DUAL_BAT_BO_DICT = {"2.4GhzOutputPower": 2, 'txPeriodMs': 75, 'rxTxPeriodMs': 255, 'energyPattern': 50}
LEGACY_BO_DICT = {'energyPattern': 17, 'rxTxPeriodMs': 99}
# BO_DICT_PREV = {'energyPattern': 36}
BO_DICT = {
    'datapath': {'config': {'pktFilter': 'Disable forwarding'}},
    'energy2400': {'config': {'energyPattern2400': 'No Energizing'}},
    'energySub1g': {'config': {'sub1gEnergyPattern': 'No Energizing'}},
    'calibration': {'config': {'calibPattern': 'Disable calibration beaconing'}}
}


# Configuration tool / change Brg/Gw Defines
ATC_GW_CONFIG = {
    'wifi': {'gwDataSrc': 'Bridges only (ch38)',
             'pacerInterval': 60,
             'txPeriodMs': 3,
             'rxTxPeriodMs': 90,
             'energizingPattern': 17,
             "2.4GhzOutputPower": 8},
    'lte': {},
    'mobile': {},
    'unknown': {}}
ATC_REGION_DICT = {
    'IL': {
        'sub1GhzFrequency': 919100,
        'energyPattern': 50
    }
}

energy_patterns = [18, 25, 26]
energy_patterns_db = {17: 50, 18: 51, 25: 56, 26: 57}
gw_rx_channel = {17: "Bridges only (ch39)", 18: "Bridges only (ch38)", 24: "Bridges only (ch39)",
                 25: "Bridges only (ch39)", 26: "Bridges only (ch39)", 51: "Bridges only (ch38)",
                 55: "Bridges only (ch39)", 56: "Bridges only (ch39)", 57: "Bridges only (ch39)",
                 50: "Bridges only (ch39)"}
shifted_brg_energy_patterns = [33, 34, 35, 58, 59, 60]
shifted_gw_energizing_patterns = [33]
lc_output_powers = [14, 17, 20, 23, 26, 29, 32]
fp_duty_cycles = [0.1, 0.15, 0.2, 0.25, 0.3]

# Power Mgmt
EXIT_POWER_MGMT_GW_DICT = {
    'gwDataSrc': gw_rx_channel[17],
    'txPeriodMs': 3,
    'rxTxPeriodMs': 15,
    'energizingPattern': 17,
    'gwMgmtMode': 'active'
}
KEEP_ALIVE_PERIOD = 30  # seconds
KEEP_ALIVE_SCAN_DURATION = 300  # in millisecond
SEC_TO_SEND = 2
BROADCAST_DST_MAC = 'FFFFFFFFFFFF'

colors = ['red', 'blue', 'yellow', 'cyan', 'green', 'brown', 'orange', 'pink', 'purple', 'black']

# Test Tool
WH_OWNER = '832742983939'
INIT_GW_CONFIG = {
    'wifi':{'gwDataSrc': gw_rx_channel[18],
            'gwDataMode': gw_rx_channel[18]},
    'lte':{},
    'mobile':{},
    'unknown':{}}
INIT_BRG_CONFIG = {
	'calibration': {'config': {'calibPattern': 'Disable calibration beaconing', 'calibInterval': 10, 'calibOutputPower': 2}},
    'externalSensor': {'config': {'adType0': 0, 'adType1': 0, 'uuidLsb0': 0, 'uuidLsb1': 0, 'uuidMsb0': 0, 'uuidMsb1': 0, 'sensor0Scramble': 0, 'sensor1Scramble': 0}},
    'energy2400': {'config': {'dutyCycle': 30, 'outputPower': 2, 'energyPattern2400': 'No Energizing'}},
    'datapath': {'config': {'pktFilter': 'Disable forwarding', 'txRepetition': 0, 'adaptivePacer': 0, 'pacerInterval': 15, 'unifiedEchoPkt': 0, 'commOutputPower': 2, 'globalPacingGroup': 0}},
    'powerManagement': {'config': {'staticLedsOn': 1, 'dynamicLedsOn': 1, 'staticOnDuration': 0, 'dynamicOnDuration': 0, 'staticKeepAliveScan': 300, 'staticSleepDuration': 0, 'dynamicKeepAliveScan': 300, 'dynamicSleepDuration': 0, 'staticKeepAlivePeriod': 20, 'dynamicKeepAlivePeriod': 20}},
    'energySub1g': {'config': {'cycle': 15, 'dutyCycle': 30, 'outputPower': 29, 'sub1gEnergyPattern': 'No Energizing'}},
    }

# prev init cfg for prev brgs version (lower then version 4.0)
INIT_BRG_PREV_CONFIG = {
	'energy2400': {'config': {'txPeriod': 5, 'rxTxPeriod': 15, 'outputPower': 2, 'energyPattern': 36}},
	'datapath': {'config': {'txRepetition': 0, 'pacerInterval': 15, 'globalPacingGroup': 0}}, 
	'energySub1g': {'config': {'frequency': 919100, 'outputPower': 32}}
    }

BRG_KEYS = ['calibPattern','calibInterval','calibOutputPower',
            'adType0','adType1','uuidLsb0','uuidLsb1','uuidMsb0','uuidMsb1','sensor0Scramble','sensor1Scramble',
            'txPeriod','rxTxPeriod','2.4GhzOutputPower','energyPattern','sub1GhzFrequency','sub1GhzOutputPower',
            'dutyCycle','outputPower','energyPattern2400','pktFilter','txRepetition','adaptivePacer','pktTypesMask',
            'pacerInterval','unifiedEchoPkt','commOutputPower','globalPacingGroup',
            'staticLedsOn','dynamicLedsOn','staticOnDuration','dynamicOnDuration','staticKeepAliveScan','staticSleepDuration',
            'dynamicKeepAliveScan','dynamicSleepDuration','staticKeepAlivePeriod','dynamicKeepAlivePeriod',
            'cycle','dutyCycle','outputPower','sub1gEnergyPattern']
GW_KEYS = ['gwDataSrc', 'gwDataMode']
GW_DATA_SRC = 'gwDataSrc'
GW_DATA_MODE = 'gwDataMode'
GW_KEYS_THIN = ['dataCoupling', 'useStaticLocation', 'gwMgmtMode']
GW_SHARED_KEYS = ['gw_2.4GhzOutputPower', 'gw_txPeriodMs', 'gw_pacerInterval', 'gw_energizingPattern',
                  'gw_rxTxPeriodMs']
TIME_COLUMNS = ['endTimestamp', 'startTimestamp', 'receivedTestConfigTimestamp']
TEST_CONFIG_COLUMNS = ['testId', 'testTimeMins', 'gatewaysIncluded', 'bridgesIncluded']



from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.interface.pkt_generator import BrgPktGeneratorNetwork
from wiliot_deployment_tools.interface.if_defines import *
from wiliot_deployment_tools.interface.packet_error import PacketError
import pkg_resources
import pandas as pd

CSV_NAME = 'packet_table.csv'
PACKET_TABLE_CSV_PATH = pkg_resources.resource_filename(__name__, CSV_NAME)

TEST_STRESS = 'stress'
TEST_COUPLING = 'coupling'
TEST_DOWNLINK = 'downlink'
TEST_UPLINK = 'uplink'
TEST_UNIFIED = 'unified'
TEST_SENSOR = 'sensor'

TESTS = [TEST_COUPLING, TEST_UPLINK, TEST_UNIFIED]
class GeneratedPacketTable:
    
    def __init__(self) -> None:
        self.brg_network = BrgPktGeneratorNetwork()
        self.table = pd.read_csv(PACKET_TABLE_CSV_PATH)
    
    def get_data(self, test, duplication, time_delay, bridge_idx) -> list:    
        assert test in TESTS, 'Invalid Test'
        assert (duplication in UPLINK_DUPLICATIONS) or (duplication in UNIFIED_DUPLICATIONS), 'Invalid Duplication'
        assert (time_delay in UPLINK_TIME_DELAYS) or (time_delay in UNIFIED_TIME_DELAYS), 'Invalid Time Delay'
        assert bridge_idx in BRIDGES, 'Invalid Bridge'
        
        t = self.table
        return t.loc[((t['test'] == test) &
                      (t['duplication']==duplication) &
                      (t['time_delay'] == time_delay) &
                      (t['bridge_idx'] == bridge_idx))].to_dict('records')[0]
            
    def get_df(self, columns=None):
        if columns is not None:
            return self.table[columns]
        return self.table
    
    def get_stress_data(self) -> pd.DataFrame:
        t = self.table
        return t.loc[((t['test'] == 'stress'))] 
    
    def get_sensor_data(self) -> pd.DataFrame:
        t = self.table
        return t.loc[((t['test'] == 'sensor'))] 
    
    def _generate_packet_table(self):
        packet_list = []
        
        # UPLINK TEST
        for duplication in UPLINK_DUPLICATIONS:
            debug_print(f'Duplication {duplication}')
            for time_delay in UPLINK_TIME_DELAYS:
                debug_print(f'Time Delay {time_delay}')
                pkts = self.brg_network.get_new_pkt_pairs()
                for idx, brg in enumerate(self.brg_network.brg_list):
                    debug_print(f'Bridge {idx}')
                    data = pkts[idx]['data_packet']
                    si = pkts[idx]['si_packet']
                    brg_id = self.brg_network.brg_list[idx].bridge_id
                    # log the sent packet with relevant info from run
                    expected_pkt = brg.get_expected_uncoupled_mqtt()
                    for pkt in expected_pkt:
                        pkt.update({'duplication': duplication, 'time_delay': time_delay})
                    packet_list.append({'test': TEST_UPLINK,
                                        'duplication': duplication,
                                        'time_delay': time_delay,
                                        'bridge_idx': idx,
                                        'expected_mqtt': expected_pkt
                                        ,'data': data, 'si': si, 'bridge_id': brg_id,
                                        })
                    
        # UNIFIED TEST (BASED ON UPLINK TEST)
        for duplication in UNIFIED_DUPLICATIONS:
            debug_print(f'Duplication {duplication}')
            for time_delay in UNIFIED_TIME_DELAYS:
                debug_print(f'Time Delay {time_delay}')
                pkts = self.brg_network.get_new_pkt_unified()
                for idx, brg in enumerate(self.brg_network.brg_list):
                    debug_print(f'Bridge {idx}')
                    data = pkts[idx]['data_packet']
                    brg_id = self.brg_network.brg_list[idx].bridge_id
                    # log the sent packet with relevant info from run
                    expected_pkt = brg.get_expected_mqtt_unified(full_data_pkt=data)
                    for pkt in expected_pkt:
                        pkt.update({'duplication': duplication, 'time_delay': time_delay})
                    packet_list.append({'test': TEST_UNIFIED,
                                        'duplication': duplication,
                                        'time_delay': time_delay,
                                        'bridge_idx': idx,
                                        'expected_mqtt': expected_pkt
                                        ,'data': data, 'bridge_id': brg_id,
                                        })        

        #STRESS TEST
        i = 0
        while i < 5000:
            i += 1
            pkts = self.brg_network.get_new_pkt_unified()
            target_idx = 0  
            brg = self.brg_network.brg_list[target_idx]
            debug_print(f'Bridge {target_idx}')
            data = pkts[target_idx]['data_packet']
            brg_id = brg.bridge_id
            expected_pkt = brg.get_expected_mqtt_unified(full_data_pkt=data)
            packet_list.append({
                'test': TEST_STRESS,
                'bridge_idx': target_idx,
                'expected_mqtt': expected_pkt,
                'data': data,
                'bridge_id': brg_id,
            })          

        def _sensor_data(df):
            hardcoded_data = [
                {"test": "sensor", "duplication": "", "time_delay": "", "bridge_idx": 2, "expected_mqtt": [{"timestamp": 17305216968476073494, "payload": "90FC020000B54D387A227038613C0D8E0FF376D199289FE6679AA0D902", "duplication": 3, "time_delay": 20}, {"timestamp": 4278768225950047766, "payload": "C6FC0000EB3B613D08177B87B44500000000000000000000009AA0D902", "duplication": 3, "time_delay": 20}], "data": "F0287B2557141E1690FC020000B54D387A227038613C0D8E0FF376D199289FE6679AA0D902", "si": "3B613D0817FB1E16C6FC0000EB3B613D08177B87B44500000000000000000000009AA0D902", "bridge_id": "3B613D08177B"},
                {"test": "sensor", "duplication": "", "time_delay": "", "bridge_idx": 2, "expected_mqtt": [{"timestamp": 17305216968476073494, "payload": "90FC0200003A79C9BEF41223AC3F4A962D97F519C9DA06CA5F4F643826", "duplication": 3, "time_delay": 20}, {"timestamp": 4278768225950047766, "payload": "C6FC0000EB3B613D08177B7D396B00000000000000000000004F643826", "duplication": 3, "time_delay": 20}], "data": "F0287B2557141E1690FC0200003A79C9BEF41223AC3F4A962D97F519C9DA06CA5F4F643826", "si": "3B613D0817FB1E16C6FC0000EB3B613D08177B7D396B00000000000000000000004F643826", "bridge_id": "3B613D08177B"},
                {"test": "sensor", "duplication": "", "time_delay": "", "bridge_idx": 2, "expected_mqtt": [{"timestamp": 17305216968476073494, "payload": "90FC0200001B65E90DDC0090175F67F3F8E7E3DA2FFA6A945076D1D979", "duplication": 3, "time_delay": 20}, {"timestamp": 4278768225950047766, "payload": "C6FC0000EB3B613D08177B39B69D000000000000000000000076D1D979", "duplication": 3, "time_delay": 20}], "data": "F0287B2557141E1690FC0200001B65E90DDC0090175F67F3F8E7E3DA2FFA6A945076D1D979", "si": "3B613D0817FB1E16C6FC0000EB3B613D08177B39B69D000000000000000000000076D1D979", "bridge_id": "3B613D08177B"},
                {"test": "sensor", "duplication": "", "time_delay": "", "bridge_idx": 2, "expected_mqtt": [{"timestamp": 17305216968476073494, "payload": "90FC020000D337363E43CDDDFC897492917BCC2DA2977E2751E31CAAA6", "duplication": 3, "time_delay": 20}, {"timestamp": 4278768225950047766, "payload": "C6FC0000EB3B613D08177B3E21640000000000000000000000E31CAAA6", "duplication": 3, "time_delay": 20}], "data": "F0287B2557141E1690FC020000D337363E43CDDDFC897492917BCC2DA2977E2751E31CAAA6", "si": "3B613D0817FB1E16C6FC0000EB3B613D08177B3E21640000000000000000000000E31CAAA6", "bridge_id": "3B613D08177B"},
                {"test": "sensor", "duplication": "", "time_delay": "", "bridge_idx": 2, "expected_mqtt": [{"timestamp": 17305216968476073494, "payload": "90FC0200001D646652892618D2CE80CA4A880A8CF46346C7743C509B54", "duplication": 3, "time_delay": 20}, {"timestamp": 4278768225950047766, "payload": "C6FC0000EB3B613D08177B5BCD2900000000000000000000003C509B54", "duplication": 3, "time_delay": 20}], "data": "F0287B2557141E1690FC0200001D646652892618D2CE80CA4A880A8CF46346C7743C509B54", "si": "3B613D0817FB1E16C6FC0000EB3B613D08177B5BCD2900000000000000000000003C509B54", "bridge_id": "3B613D08177B"},
                {"test": "sensor", "duplication": "", "time_delay": "", "bridge_idx": 2, "expected_mqtt": [{"timestamp": 17305216968476073494, "payload": "90FC020000CEE15F2E2EE026B22C739867453C6E2E108E9E8610FE7F07", "duplication": 3, "time_delay": 20}, {"timestamp": 4278768225950047766, "payload": "C6FC0000EB3B613D08177B9647FC000000000000000000000010FE7F07", "duplication": 3, "time_delay": 20}], "data": "F0287B2557141E1690FC020000CEE15F2E2EE026B22C739867453C6E2E108E9E8610FE7F07", "si": "3B613D0817FB1E16C6FC0000EB3B613D08177B9647FC000000000000000000000010FE7F07", "bridge_id": "3B613D08177B"},
                {"test": "sensor", "duplication": "", "time_delay": "", "bridge_idx": 2, "expected_mqtt": [{"timestamp": 17305216968476073494, "payload": "90FC0200002CC60917A02CB02367771BE9EA20F9666ED8A06F6612745B", "duplication": 3, "time_delay": 20}, {"timestamp": 4278768225950047766, "payload": "C6FC0000EB3B613D08177BF1EBC500000000000000000000006612745B", "duplication": 3, "time_delay": 20}], "data": "F0287B2557141E1690FC0200002CC60917A02CB02367771BE9EA20F9666ED8A06F6612745B", "si": "3B613D0817FB1E16C6FC0000EB3B613D08177BF1EBC500000000000000000000006612745B", "bridge_id": "3B613D08177B"},
                {"test": "sensor", "duplication": "", "time_delay": "", "bridge_idx": 2, "expected_mqtt": [{"timestamp": 17305216968476073494, "payload": "90FC02000007EFE229109B044DB995A506179C99094720AE8BF9F78C1A", "duplication": 3, "time_delay": 20}, {"timestamp": 4278768225950047766, "payload": "C6FC0000EB3B613D08177B6CEC9A0000000000000000000000F9F78C1A", "duplication": 3, "time_delay": 20}], "data": "F0287B2557141E1690FC02000007EFE229109B044DB995A506179C99094720AE8BF9F78C1A", "si": "3B613D0817FB1E16C6FC0000EB3B613D08177B6CEC9A0000000000000000000000F9F78C1A", "bridge_id": "3B613D08177B"},
                {"test": "sensor", "duplication": "", "time_delay": "", "bridge_idx": 2, "expected_mqtt": [{"timestamp": 17305216968476073494, "payload": "90FC0200002CA1E008E364D0DCF65631718BCD659DE3323A69A674F7B9", "duplication": 3, "time_delay": 20}, {"timestamp": 4278768225950047766, "payload": "C6FC0000EB3B613D08177B1051900000000000000000000000A674F7B9", "duplication": 3, "time_delay": 20}], "data": "F0287B2557141E1690FC0200002CA1E008E364D0DCF65631718BCD659DE3323A69A674F7B9", "si": "3B613D0817FB1E16C6FC0000EB3B613D08177B1051900000000000000000000000A674F7B9", "bridge_id": "3B613D08177B"},
                {"test": "sensor", "duplication": "", "time_delay": "", "bridge_idx": 2, "expected_mqtt": [{"timestamp": 17305216968476073494, "payload": "90FC02000049ADD722F535679C37983927655C974A4980B080045DA6C2", "duplication": 3, "time_delay": 20}, {"timestamp": 4278768225950047766, "payload": "C6FC0000EB3B613D08177B5280B20000000000000000000000045DA6C2", "duplication": 3, "time_delay": 20}], "data": "F0287B2557141E1690FC02000049ADD722F535679C37983927655C974A4980B080045DA6C2", "si": "3B613D0817FB1E16C6FC0000EB3B613D08177B5280B20000000000000000000000045DA6C2", "bridge_id": "3B613D08177B"},
                {"test": "sensor", "duplication": "", "time_delay": "", "bridge_idx": 2, "expected_mqtt": [{"timestamp": 17305216968476073494, "payload": "90FC0200004C8AD49C6D96923BFB70DF06554F5E8F438F1DF57E063773", "duplication": 3, "time_delay": 20}, {"timestamp": 4278768225950047766, "payload": "C6FC0000EB3B613D08177BB4085900000000000000000000007E063773", "duplication": 3, "time_delay": 20}], "data": "F0287B2557141E1690FC0200004C8AD49C6D96923BFB70DF06554F5E8F438F1DF57E063773", "si": "3B613D0817FB1E16C6FC0000EB3B613D08177BB4085900000000000000000000007E063773", "bridge_id": "3B613D08177B"},
                {"test": "sensor", "duplication": "", "time_delay": "", "bridge_idx": 2, "expected_mqtt": [{"timestamp": 17305216968476073494, "payload": "90FC02000064A5B54285BB6BCDB457ABBED8EE26B4EB43B27A8C26781C", "duplication": 3, "time_delay": 20}, {"timestamp": 4278768225950047766, "payload": "C6FC0000EB3B613D08177B27363800000000000000000000008C26781C", "duplication": 3, "time_delay": 20}], "data": "F0287B2557141E1690FC02000064A5B54285BB6BCDB457ABBED8EE26B4EB43B27A8C26781C", "si": "3B613D0817FB1E16C6FC0000EB3B613D08177B27363800000000000000000000008C26781C", "bridge_id": "3B613D08177B"}
            ]
        
            hardcoded_df = pd.DataFrame(hardcoded_data)
            return pd.concat([df, hardcoded_df], ignore_index=True)
        
        df = pd.DataFrame(packet_list)
        df = _sensor_data(df)
        df.to_csv(PACKET_TABLE_CSV_PATH)

class CouplingRunData:
    def __init__(self, data) -> None:
        self.test = data['test']
        self.duplication = data['duplication']
        self.time_delay = data['time_delay']
        self.bridge_idx = data['bridge_idx']
        self.packet_error = eval(data['packet_error'])
        self.expected_mqtt = eval(data['expected_mqtt'])
        self.data = data['data']
        self.si = data['si']
        self.bridge_id = data['bridge_id']
        self.scattered_time_delay = data['scattered_time_delay']


    @classmethod
    def get_data(cls, test, duplication, time_delay, bridge_idx):
        packet_data = GeneratedPacketTable().get_data(test, duplication, time_delay, bridge_idx)
        return cls(packet_data)

class UplinkRunData:
    def __init__(self, data) -> None:
        self.test = data['test']
        self.duplication = data['duplication']
        self.time_delay = data['time_delay']
        self.bridge_idx = data['bridge_idx']
        self.expected_mqtt = eval(data['expected_mqtt'])
        self.data = data['data']
        self.si = data['si']
        self.bridge_id = data['bridge_id']   

    @classmethod
    def get_data(cls, test, duplication, time_delay, bridge_idx):
        packet_data = GeneratedPacketTable().get_data(test, duplication, time_delay, bridge_idx)
        return cls(packet_data)
    
class UnifiedRunData:
    def __init__(self, data) -> None:
        self.test = data['test']
        self.duplication = data['duplication']
        self.time_delay = data['time_delay']
        self.bridge_idx = data['bridge_idx']
        self.expected_mqtt = eval(data['expected_mqtt'])
        self.data = data['data']
        self.bridge_id = data['bridge_id']   

    @classmethod
    def get_data(cls, test, duplication, time_delay, bridge_idx):
        packet_data = GeneratedPacketTable().get_data(test, duplication, time_delay, bridge_idx)
        return cls(packet_data)
    
    @classmethod
    def get_df(cls, columns=None):
       df = GeneratedPacketTable().get_df(columns)
       return df[df['test'] == TEST_UNIFIED]


class StressRunData:
    def __init__(self) -> None:
        self.data = GeneratedPacketTable().get_stress_data()

class SensorRunData:
    def __init__(self) -> None:
        self.data = GeneratedPacketTable().get_sensor_data()
        
if __name__ == "__main__":
    GeneratedPacketTable()._generate_packet_table()



# Files
UT_MQTT_LOG_FILE = "ut_mqtt_log.json"

# GW defines
GW_ID =                         "gatewayId"
ADDITIONAL =                    "additional"
REPORTED_CONF =                 "reportedConf"
GW_CONF =                       "gatewayConf"
GW_NAME =                       "gatewayName"
GW_API_VERSION =                "apiVersion"
LAT =                           "lat"
LNG =                           "lng"
WLT_SERVER =                    "wiliotServer"
PACER_INTERVAL =                "pacerInterval"
PKT_TYPES_MASK =                "packetTypesMask"
RX_TX_PERIOD =                  "rxTxPeriodMs"
TX_PERIOD =                     "txPeriodMs"
ENERGY_PATTERN =                "energyPattern"
OUTPUT_POWER_2_4 =              "2.4GhzOutputPower"
NFPKT =                         "nfpkt"
RSSI =                          "rssi"
SRC_ID =                        "src_id"

MUL_BRG_TEST_MIN =              2
INTERNAL_BRG_RSSI =             1
BRIDGE_ID =                     "bridgeId"
ALIAS_BRIDGE_ID =               "aliasBridgeId"
GROUP_ID =                      "group_id"
BRG_ACTION =                    "bridgeAction"
INTERNAL_BRG_STR =              "internal_brg"
INTERNAL =                      "INTERNAL"

GW_DATA_MODE =                  "gwDataMode"
TAGS_AND_BRGS =                 "Tags & Bridges"
TAGS_ONLY =                     "Tags only"
BRGS_ONLY_37 =                  "Bridges only (ch37)"
BRGS_ONLY_38 =                  "Bridges only (ch38)"
BRGS_ONLY_39 =                  "Bridges only (ch39)"
HIBERNATE =                     "Hibernate"

BLE_WIFI =                      "ble_wifi"
BLE_LAN =                       "ble_lan"

WLT_SERVER =                    "wiliotServer"
PACER_INTERVAL =                "pacerInterval"
OUTPUT_POWER_2_4 =              "2.4GhzOutputPower"
USE_STAT_LOC =                  "useStaticLocation"
GW_ENERGY_PATTERN =             "energizingPattern"
VERSION =                       "version"
WIFI_VERSION =                  "interfaceChipSwVersion"
BLE_VERSION =                   "bleChipSwVersion"
DATA_COUPLING =                 "dataCoupling"
GW_MODE =                       "gwMode"
PROD =                          "prod"

GET_INFO_ACTION =               "getGwInfo"
REBOOT_GW_ACTION =              "rebootGw"
LOG_PERIOD_ACTION =             "LogPeriodSet"
GET_LOGS =                      "getLogs"
GW_INFO =                       "gatewayInfo"
GW_LOGS =                       "gatewayLogs"
GW_LATITUDE =                   "Latitude"
GW_LONGITUDE =                  "Longitude"
GW_LOG_PERIOD =                 30

# Thin gw defines
THIN_GW_PROTOCOL_VERSION =      "protocolVersion"
TX_PKT =                        "txPacket"
TX_MAX_DURATION_MS =            "txMaxDurationMs"
TX_MAX_RETRIES =                "txMaxRetries"
TRANPARENT_PKT_LEN =            31 * 2
COUPLED_DATA_PKT_LEN =          29 * 2 # No 1E16

# Configurable brg fields' names
BRG_TX_REPETITION =             "tx_repetition"
BRG_GLOBAL_PACING_GROUP =       "global_pacing_group"
BRG_OUTPUT_POWER =              "output_power"
BRG_CALIB_OUTPUT_POWER =        "calib_output_power"
BRG_COMM_OUTPUT_POWER =         "comm_output_power"
BRG_CALIB_INTERVAL =            "calib_interval"
BRG_CALIB_PATTERN =             "calib_pattern"
BRG_RXTX_PERIOD =               "rx_tx_period" # rx_tx_period_ms in Brg2GwCfg & Gw2BrgCfg
BRG_TX_PERIOD =                 "tx_period" # tx_period_ms in Brg2GwCfg & Gw2BrgCfg
BRG_DUTY_CYCLE =                "duty_cycle"
BRG_CYCLE =                     "cycle"
BRG_PACER_INTERVAL =            "pacer_interval"
BRG_PKT_FILTER =                "pkt_filter"
BRG_ENERGY_PATTERN_2_4 =        "energy_pattern_2400"
BRG_SUB1GHZ_ENERGY_PATTERN =    "sub1g_energy_pattern"
BRG_ADAPTIVE_PACER =            "adaptive_pacer"
BRG_UNIFIED_ECHO_PKT =          "unified_echo_pkt"
BRG_COMM_PATTERN =              "comm_pattern"

# Common defines
PACKETS =                       "packets"
TIMESTAMP =                     "timestamp"
ACTION =                        "action"
PAYLOAD =                       "payload"
SEQUENCE_ID =                   "sequenceId"
MODULE_IF =                     "module IF"
HB =                            "HB"
DATETIME =                      "datetime"
TIME =                          "time"
TIMESTAMP_DELTA =               "timestamp_delta"
TAGS_COUNT =                    "tags_count"
NEW_TAGS =                      "new_tags"
TTFP =                          "ttfp"

# External Sensors
IS_SENSOR =                      "isSensor"
IS_EMBEDDED =                    "isEmbedded"
IS_SCRAMBLED =                   "isScrambled"
SENSOR_UUID =                    "sensorServiceId"
SENSOR_ID =                      "sensorId"

# OTA
STATUS_CODE_STR =                "statusCode"
IMG_DIR_URL =                    "imageDirUrl"
UPGRADE_BLSD =                   "upgradeBlSd"
VER_UUID_STR =                   "versionUUID"
VER_MAX_LEN =                    31

# Versions
VERSIONS = {
    "1.5.0" : {WIFI_VERSION: "3.5.32", BLE_VERSION: "3.7.25"},
    "1.5.2" : {WIFI_VERSION: "3.5.132", BLE_VERSION: "3.7.25"},
    "1.6.1" : {WIFI_VERSION: "3.5.51", BLE_VERSION: "3.8.18"},
    "1.7.0" : {WIFI_VERSION: "3.9.8", BLE_VERSION: "3.9.24"},
    "1.7.1" : {WIFI_VERSION: "3.10.6", BLE_VERSION: "3.10.13"},
    "1.8.0" : {WIFI_VERSION: "3.11.36", BLE_VERSION: "3.11.40"},
    "1.8.2" : {WIFI_VERSION: "3.11.36", BLE_VERSION: "3.11.42"},
    "1.9.0" : {WIFI_VERSION: "3.12.10", BLE_VERSION: "3.12.36"},
    "1.10.1" : {WIFI_VERSION: "3.13.29", BLE_VERSION: "3.13.25"},
    "3.14.0" : {WIFI_VERSION: "3.14.33", BLE_VERSION: "3.14.64"},
    "3.15.0" : {WIFI_VERSION: "3.15.38", BLE_VERSION: "3.15.72"},
    "3.16.3" : {WIFI_VERSION: "3.16.20", BLE_VERSION: "3.16.96"},
    "3.17.0" : {WIFI_VERSION: "3.17.25", BLE_VERSION: "3.17.90"},
    "4.0.0" : {WIFI_VERSION: "4.0.8", BLE_VERSION: "4.0.65"},
    "4.1.0" : {WIFI_VERSION: "4.1.8", BLE_VERSION: "4.1.33"}
}

# Tests defines
DEFAULT_GW_FIELD_UPDATE_TIMEOUT =   10
DEFAULT_BRG_FIELD_UPDATE_TIMEOUT =  10
HB_PERIOD =                         30
VER_UPDATE_TIMEOUT =                300
GW_LATITUDE_DEFAULT =               33.0222
GW_LONGITUDE_DEFAULT =              -117.0839
# Set to work with default when versions tests only pass through new api ver
GW_API_VER_DEFAULT =                "201"
GW_API_VER_OLD =                    "200"
BRG_CFG_HAS_LEN =                   2
CLEAR_DATA_PATH_TIMEOUT =           10
BRG_ADVERTISEMENT_TIMEOUT =         30 + 2 # First 30 for wlt app start & 2 extra for brg to settle to recieve its get module action
BRG_OTA_TIMEOUT =                   100
ACTION_LONG_TIMEOUT =               120
ACTION_SI_PKT_TIMEOUT =             10
ACTION_SHORT_TIMEOUT =              5

# Internal python ut defines - used only in ut
PACER_INTERVAL_THRESHOLD_HIGH =                     0.90
PACER_INTERVAL_THRESHOLD =                          0.80
GLOBAL_PACING_GROUP_THRESHOLD =                     0.70
PACKETS_ECHO_OFF =                                  16
TEST_PASSED =                                       0
TEST_FAILED =                                       -1
NO_RESPONSE =                                       "NO_RESPONSE"
DONE =                                              "DONE"
TEST_SUCCESS =                                      ":)"
MGMT_PKT =                                          "mgmt_pkt"
UNIFIED_PKT =                                       "unified_pkt"
SIDE_INFO_PKT =                                     "side_info_pkt"
DECODED_DATA =                                      "decoded_data"
TAG_ID =                                            "tag_id"
PACKET_CNTR =                                       "packet_cntr"
PACKET_TYPE =                                       "packet_type"
PACKET_DATA =                                       "packet_data"
PKTS =                                              "pkts"
MQTT_LOG_PRE_STR =                                  "mqtt_log_"
GW_DATA =                                           "gw_data"
GW_ID =                                             "gw_id"
CER =                                               "cer"
PKT_CNTR_DIFF =                                     "packet_cntr_diff"
AVG =                                               "avg_"
CER_PER_TAG =                                       "cer_per_tag"
AWS =                                               "aws"
TEST =                                              "test"

# test reasons
BRG_VER_SUCCESS =           "SUCCESS - BRG version matches expected version!"
WANTED_VER_SAME =           "Wanted version is same as original one!"
WANTED_VER_SAME_MUL =       "Wanted versions is same as original ones!"
VER_UPDATE_PASSED =         "Version Update Ran Successfully!"
VER_UPDATE_FAILED =         "The Update Process Has Been Interrupted!"

# Non Default defines
BRG_NON_DEFAULT_DUTY_CYCLE =                        15
BRG_NON_DEFAULT_OP_2_4 =                            6
BRG_NON_DEFAULT_EP_2_4 =                            1
BRG_NON_DEFAULT_OUTPUT_POWER_SUB1G =                26
BRG_NON_DEFAULT_CYCLE_SUB1G =                       40
BRG_NON_DEFAULT_PWR_MGMT_KEEP_ALIVE_SCAN =          0
BRG_NON_DEFAULT_TX_REPETITION =                     2
BRG_NON_DEFAULT_PACER_INTERVAL =                    20
BRG_NON_DEFAULT_CALIB_OUTPUT_POWER =                8
BRG_NON_DEFAULT_PKT_FILTER =                        17
BRG_NON_DEFAULT_CALIB_PATTERN =                     2
BRG_NON_DEFAULT_CALIB_INTERVAL =                    15

LIS2DW12_NON_DEFAULT_STATE_THRESHOLD =              93
LIS2DW12_NON_DEFAULT_WAKE_UP_DURATION =             120
LIS2DW12_NON_DEFAULT_SLEEP_DURATION =               35

# ---------------------------------------------------RTSA defines---------------------------------------------------
# common defines
TRACE_LOG_FILE_NAME =                   "TRACELOG"
TRACE_LOG_FILE_PATH =                   "C:/SignalVu-PC Files/" + TRACE_LOG_FILE_NAME + ".TOV"

# freq defines
FREQ_2_4_GHZ =                          {'37':2.402, '38':2.426, '39':2.480}
FREQ_SUB1G_MHZ =                        {'865_7':865.700, '915':915.000, '916_3':916.300, '917_5':917.500, '918':918.000, '919_1':919.100}

# SignalVu API commands defines
TRACE_DETECTION =                       {'average':'AVERage', 'positive':'POSitive', 'negative':'NEGative', 'positive-negative':'POSNegative', 'sample':'SAMPle'}
MAX_TRACE_POINTS =                      {'1K':'ONEK', '10K':'TENK', '100K':'HUNDredk', 'never_decimate':'NEVerdecimate' }

# default values
DEFAULT_LENGTH_MS =                     30
DEFAULT_TIME_PER_DIVISION_SEC =         5
DEFAULT_RX_TX_PERIOD_SEC =              0.015
BEACON_MIN_LENGTH_SEC =                 375e-6
BEACON_MAX_LENGTH_SEC =                 500e-6
ENERGIZING_TIME_THRESHOLD =             0.3
BEACON_POWER_THRESHOLD =                0.9
BEACON_POWER_CURVE_38 =                 0.7
BEACON_POWER_CURVE_39 =                 0.625
DEFAULT_SPAN_MHZ =                      5
RXTX_MAX_CFG =                          255
RXTX_CFG_DEFAULT =                      100

# test times
FREQ_BEACONS_ANALYSIS_TIME_DELTA =      10

# structured energizing patterns information
class energizingPattern:
    def __init__(self, ble_calibration_beacons = [], ble_energy = {}, ble_post_energy_beacons = [], sub1G_energy = False, info = ""):
        self.ble_calibration_beacons = ble_calibration_beacons
        self.ble_energy = ble_energy
        self.ble_post_energy_beacons = ble_post_energy_beacons
        self.sub1G_energy = sub1G_energy
        self.info = info

EP_INFO = {
        '17' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']]),
        '18' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['39'] : 1.0}),
        '20' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['37'] : 0.2, FREQ_2_4_GHZ['39'] : 0.8}),
        '24' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['37'] : 1.0}),
        '25' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['38'] : 1.0}),
        '26' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={2.454 : 1.0}),
        '27' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39'], FREQ_2_4_GHZ['39']]),
        '29' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['38'] : 0.3333, 2.454 : 0.3333, FREQ_2_4_GHZ['39'] : 0.3333}),
        '36' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'],FREQ_2_4_GHZ['38'],FREQ_2_4_GHZ['39']], info="idle"),
        '37' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], 2.415, FREQ_2_4_GHZ['39'], 2.441, 2.428, 2.454, 2.467], ble_energy={2.450 : 1.0}, info="euro"),
        '50' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], sub1G_energy=True),
        '51' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['39'] : 1.0}, sub1G_energy=True),
        '52' : energizingPattern(sub1G_energy=True),
        '55' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['37'] : 1.0}, sub1G_energy=True),
        '56' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={FREQ_2_4_GHZ['38'] : 1.0}, sub1G_energy=True),
        '57' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], FREQ_2_4_GHZ['38'], FREQ_2_4_GHZ['39']], ble_energy={2.454 : 1.0}, sub1G_energy=True),
        '61' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], 2.415, FREQ_2_4_GHZ['39'], 2.441, 2.428, 2.454, 2.467], sub1G_energy=True, info="euro"),
        '62' : energizingPattern(ble_calibration_beacons=[FREQ_2_4_GHZ['37'], 2.415, FREQ_2_4_GHZ['39'], 2.441, 2.428, 2.454, 2.467], ble_energy={2.475 : 1.0}, sub1G_energy=True, info="euro"),
        '99' : energizingPattern(ble_calibration_beacons=[i/1000.0 for i in range(2402, 2481, 2)])
                }

EP_FREQ_BREAKDOWN_COUNTER_SETUP = {
    2.402 : {'beacons':0, 'energy_in_ms': 0.0},
    2.426 : {'beacons':0, 'energy_in_ms': 0.0},
    2.480 : {'beacons':0, 'energy_in_ms': 0.0},
    2.403 : {'beacons':0, 'energy_in_ms': 0.0},
    2.427 : {'beacons':0, 'energy_in_ms': 0.0},
    2.483 : {'beacons':0, 'energy_in_ms': 0.0},
    2.454 : {'beacons':0, 'energy_in_ms': 0.0},
    2.481 : {'beacons':0, 'energy_in_ms': 0.0},
    2.415 : {'beacons':0, 'energy_in_ms': 0.0},
    2.441 : {'beacons':0, 'energy_in_ms': 0.0},
    2.428 : {'beacons':0, 'energy_in_ms': 0.0},
    2.467 : {'beacons':0, 'energy_in_ms': 0.0},
    2.475 : {'beacons':0, 'energy_in_ms': 0.0},
    0.8657 : {'beacons':0, 'energy_in_ms': 0.0},
    0.915 : {'beacons':0, 'energy_in_ms': 0.0},
    0.9163 : {'beacons':0, 'energy_in_ms': 0.0},
    0.9175 : {'beacons':0, 'energy_in_ms': 0.0},
    0.918 : {'beacons':0, 'energy_in_ms': 0.0},
    0.9191 : {'beacons':0, 'energy_in_ms': 0.0}
}

# Tag IDs for setup 1 & 4 - private setups in a chamber
# Added to support tag count in versions_analysis_test
NO_TAGS_STR = "no_tags"
SETUP_1_STR = "setup_1"
SETUP_4_STR = "setup_4"

PRIVATE_SETUP_CHOICES = [NO_TAGS_STR,
                         SETUP_1_STR,
                         SETUP_4_STR]

SETUP_1_TAG_IDS = [
    "000d40174823",
    "000d401747df",
    "000d401747f2",
    "000d40174862",
    "000d4017483d",
    "000d40174843",
    "000d401747f1",
    "000d4017482a",
    "000d401747f7",
    "000d40174869",
    "000d40174824",
    "000d40174829",
    "000d40174810",
    "000d4017480b",
    "000d401747d1",
    "000d4017480a",
    "000d401747f8",
    "000d401747cc",
    "000d40174882",
    "000d40174863",
    "000d4017483b",
    "000d401747cb",
    "000d40174842",
    "000d4017487c",
    "000d40174811",
    "000d401747e0",
    "000d40174881"
]
SETUP_4_TAG_IDS = [
    "000d40ce912e",
    "000d40ce9107",
    "000d40ce9233",
    "000d40ce9240",
    "000d40ce9219",
    "000d40ce912d",
    "000d40ce9120",
    "000d40ce9228",
    "000d40ce9146",
    "000d40ce9138",
    "000d40ce9200",
    "000d40ce9229",
    "000d40ce921a",
    "000d40ce9148",
    "000d40ce911f",
    "000d40ce920e",
    "000d40ce9232",
    "000d40ce9201",
    "000d40ce9137"
]

TAG_IDS = {
    SETUP_1_STR: SETUP_1_TAG_IDS,
    SETUP_4_STR: SETUP_4_TAG_IDS
}


CSV_NAME = 'bad_crc_to_PER_quantization.csv'
INTERFERENCE_ANALYSIS_FW_VER = "4.2.0"
CHANNELS_TO_ANALYZE = [(37, 2402), (38, 2426), (39, 2480)]
CNTRS_LISTEN_TIME_SEC = 30
MAX_UNSIGNED_32_BIT = 4294967295
NON_WLT_RX = 'non_wlt_rx'
WLT_RX = 'wlt_rx'
BAD_CRC = 'bad_crc'
CNTRS_KEYS = [NON_WLT_RX, WLT_RX, BAD_CRC]


from wiliot_deployment_tools.ag.ut_defines import *
from wiliot_deployment_tools.ag.wlt_types_ag import *
def eval_pkt(str):
    try:
        return eval(str)
    except:
        return None

class WltPkt():
    supported_pkt_types = WLT_PKT_TYPES

    def __init__(self, raw='', hdr=None, data_hdr=None, generic=None, pkt=None):
        self.hdr = Hdr() if hdr is None else hdr
        self.data_hdr = DataHdr() if data_hdr is None else data_hdr
        self.generic = generic
        self.pkt = pkt
        if raw:
            self.set(raw)

    def __eq__(self, other):
        if isinstance(other, WltPkt):
            return (
                (self.hdr == other.hdr or
                self.data_hdr == other.data_hdr) and
                self.generic == other.generic and
                self.pkt == other.pkt
            )
        return False

    def dump(self):
        if self.pkt:
            if isinstance(self.pkt, UnifiedEchoPkt):
                return self.data_hdr.dump() + self.pkt.dump()
            else:
                return self.hdr.dump() + self.pkt.dump()
        return self.data_hdr.dump() + self.generic.dump()
    
    def set(self, string):
        if not string.startswith("1E16"):
            string = "1E16" + string

        self.hdr.set(string[0:14])
        self.data_hdr.set(string[0:14])
        if self.hdr.group_id == GROUP_ID_BRG2GW or self.hdr.group_id == GROUP_ID_GW2BRG:
            # GROUP_ID_BRG2GW & GROUP_ID_GW2BRG
            self.generic = eval_pkt(f'GenericV{API_VERSION_LATEST}()')
            self.generic.set(string[14:62])

            # MEL modules
            if self.generic.module_type:
                if self.generic.module_type == MODULE_CUSTOM:
                    self.pkt = eval_pkt(f'ModuleCustomLis2Dw12V{self.generic.api_version}()')
                else:
                    self.pkt = eval_pkt(f'{MODULES_DICT[self.generic.module_type]}{self.generic.api_version}()')
            elif self.generic.module_type == MODULE_GLOBAL:
                # Action pkts
                if self.generic.msg_type == BRG_MGMT_MSG_TYPE_ACTION:
                    pkt = eval_pkt(f'ActionGenericV{self.generic.api_version}("{string[14:62]}")')
                    if self.generic.api_version >= API_VERSION_V8 and (type(pkt) != type(None)):
                        pkt = eval_pkt(f'{ACTIONS_DICT[pkt.action_id]}{self.generic.api_version}()')
                    self.pkt = pkt
                # OLD global config (GW2BRG & BRG2GW)
                elif self.hdr.group_id == GROUP_ID_GW2BRG and self.generic.msg_type == BRG_MGMT_MSG_TYPE_CFG_SET:
                    self.pkt = eval_pkt(f'Gw2BrgCfgV8()') # no api_version field in Gw2BrgCfg pkts - default parse as api version 8 (Latest Gw2BrgCfg)
                elif self.hdr.group_id == GROUP_ID_BRG2GW:
                    if self.generic.msg_type == BRG_MGMT_MSG_TYPE_CFG_SET or self.generic.msg_type == BRG_MGMT_MSG_TYPE_CFG_INFO:
                        self.pkt = eval_pkt(f'Brg2GwCfgV{self.generic.api_version}()')
                # Brg2GwHb pkts
                    elif self.generic.msg_type == BRG_MGMT_MSG_TYPE_HB:
                        self.pkt = eval_pkt(f'Brg2GwHbV{self.generic.api_version}()')
        # Unified pkt
        elif self.data_hdr.group_id_major == GROUP_ID_UNIFIED_PKT:
            self.pkt = UnifiedEchoPkt()
        # SideInfo pkts
        elif self.hdr.group_id == GROUP_ID_SIDE_INFO_SENSOR:
            self.pkt = SideInfoSensor()
        elif self.hdr.group_id == GROUP_ID_SIDE_INFO:
            self.pkt = SideInfo()

        if self.pkt:
            self.pkt.set(string[14:62])

hex_str2int = lambda x: int(x,16)


import copy
import datetime
import os
import re
import time
from typing import Literal
import pandas as pd
import plotly.express as px
import tabulate
from wiliot_deployment_tools.ag.ut_defines import PAYLOAD
from wiliot_deployment_tools.api.extended_api import GatewayType
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.gw_certificate.api_if.gw_capabilities import GWCapabilities
from wiliot_deployment_tools.interface.ble_simulator import BLESimulator
from wiliot_deployment_tools.interface.uart_if import UARTInterface
from wiliot_deployment_tools.gw_certificate.tests.static.uplink_defines import *
from wiliot_deployment_tools.interface.mqtt import MqttClient
from wiliot_deployment_tools.interface.pkt_generator import BrgPktGenerator
from wiliot_deployment_tools.gw_certificate.tests.static.generated_packet_table import StressRunData
from wiliot_deployment_tools.gw_certificate.tests.generic import PassCriteria, PERFECT_SCORE, GenericTest, GenericStage, INFORMATIVE
from wiliot_deployment_tools.interface.packet_error import PacketError
from wiliot_deployment_tools.gw_certificate.tests.static.generated_packet_table import CSV_NAME
import pkg_resources


# HELPER FUNCTIONS

def process_payload(packet:dict):
    payload = packet[PAYLOAD]
    payload = payload.upper()
    if len(payload) == 62:
        if payload[:4] == '1E16':
            payload = payload [4:]
    # big2little endian
    if payload[:4] == 'FCC6':
        payload = 'C6FC' + payload[4:]
    packet[PAYLOAD] = payload
    return packet


# TEST STAGES

class StressTestError(Exception):
    pass

class GenericStressStage(GenericStage):
    def __init__(self, mqttc:MqttClient, ble_sim:BLESimulator, gw_capabilities:GWCapabilities, stage_name,
                **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(stage_name=stage_name, **self.__dict__)
        
        self.result_indication = INFORMATIVE        
        
        # Clients
        self.mqttc = mqttc
        self.ble_sim = ble_sim

        # Packets list
        self.local_pkts = []
        self.mqtt_pkts = []
        self.full_test_pkts = pd.DataFrame()
        
        # GW Capabilities
        self.gw_capabilities = gw_capabilities
        
        # Packet Error
        self.packet_error = PacketError()

        #Run Data
        self.run_stress_data = StressRunData()
        
        # Data extracted from the test csv
        self.all_test_payloads = None
    
    def prepare_stage(self, reset_ble_sim=True):
        super().prepare_stage()
        self.mqttc.flush_messages()
        if reset_ble_sim:
            self.ble_sim.set_sim_mode(True) 
        
    def fetch_mqtt_from_stage(self):
        mqtt_pkts = self.mqttc.get_all_tags_pkts()
        self.mqtt_pkts = list(map(lambda p: process_payload(p), mqtt_pkts))
    
    def compare_local_mqtt(self):
        self.fetch_mqtt_from_stage()
        local_pkts_df = pd.DataFrame(self.local_pkts)
        mqtt_pkts_df = pd.DataFrame(self.mqtt_pkts)
        comparison = local_pkts_df

        if PAYLOAD not in mqtt_pkts_df.columns:
            mqtt_pkts_df[PAYLOAD] = ''
        received_pkts_df = pd.merge(local_pkts_df[PAYLOAD], mqtt_pkts_df[PAYLOAD], how='inner')
        
        received_pkts = set(received_pkts_df[PAYLOAD])

        self.pkts_received_count = pd.Series.count(received_pkts_df)
        unique_received_count = len(received_pkts)
        self.pkts_filtered_out_count = self.pkts_received_count - unique_received_count

        comparison[RECEIVED] = comparison[PAYLOAD].isin(received_pkts)
        comparison['pkt_id'] = comparison['payload'].apply(lambda x: x[-8:])
        self.comparison = comparison

class StressTestStage(GenericStressStage):
    
    def __init__(self, mqttc, ble_sim, gw_capabilities, **kwargs):
        super().__init__(mqttc, ble_sim, gw_capabilities, stage_name=type(self).__name__, **kwargs)
        self.delays = [255, 200, 130, 100, 60, 30, 20, 10, 5]
        self.duplicates = 3
        self.report_data = {}
        self.stage_tooltip = "Attempts different PPS (packets per second) rates to examine the gateway limit"
        
    def run(self):
        for delay in self.delays:
            self.stress_delay = delay
            debug_print(f"---Running Stress Test with delay of {delay}")
            self.run_full_stress_test(delay)

    def run_full_stress_test(self, delay):
        self.prepare_stage()
        self.run_stress_test_with_delay(delay)
        self.generate_stage_report()
        self.teardown_stage()

    def prepare_stage(self):
        super().prepare_stage()
        self.local_pkts = []
        self.start_time = datetime.datetime.now()

    def run_stress_test_with_delay(self, delay):
        run_data = self.run_stress_data.data
        for index, row in run_data.iterrows():
            if datetime.datetime.now() > self.start_time + datetime.timedelta(seconds=60):
                break
            data = row['data']
            expected_mqtt = eval(row['expected_mqtt'])
            self.local_pkts.extend(expected_mqtt)
            self.ble_sim.send_packet(data, duplicates=self.duplicates, delay=delay)
        time.sleep(15)

    def teardown_stage(self):
        self.ble_sim.set_sim_mode(False)
        self.mqttc.flush_messages()

    def generate_stage_report(self):
        def save_to_csv(df, csv_path):
            """
            Save a DataFrame to a CSV file without overwriting existing content.
            
            :param df: The DataFrame to be saved.
            :param csv_path: The path to the CSV file.
            """
            if os.path.exists(csv_path) and os.path.getsize(csv_path) > 0:
                df.to_csv(csv_path, mode='a', header=False, index=False)
            else:
                df.to_csv(csv_path, mode='w', header=True, index=False)
                
        self.compare_local_mqtt()
        report = []
        num_pkts_sent = len(self.comparison)
        num_pkts_received = self.comparison['received'].eq(True).sum()
        self.stage_pass = PERFECT_SCORE    
        self.comparison['time_delay'] = self.stress_delay
        self.full_test_pkts = pd.concat([self.full_test_pkts, self.comparison], ignore_index=True)
        save_to_csv(self.comparison, self.csv_path)
        
        if str(self.stress_delay) not in self.report_data:
                self.report_data[str(self.stress_delay)] = {}
        self.report_data[str(self.stress_delay)]['pkts_sent'] = num_pkts_sent
        self.report_data[str(self.stress_delay)]['pkts_recieved'] = num_pkts_received
        self.report_data[str(self.stress_delay)]['pkts_per_sec_desired'] = 1000 / (self.stress_delay * self.duplicates)
        self.report_data[str(self.stress_delay)]['pkts_per_sec_actual'] = num_pkts_sent / 60
        self.report_data[str(self.stress_delay)]['percent_of_false_recieved'] = 100 - (num_pkts_received * 100 / num_pkts_sent)
        
        if self.stress_delay == self.delays[-1]:
            
            self.stage_pass = PassCriteria.calc_for_stage_stress(self.stage_pass, self.stage_name)
            
            for delay in self.delays:
                delay = str(delay)
                report.append((f'Test - {self.report_data[delay]["pkts_per_sec_desired"]} packets per second. Actual packet per second - ', self.report_data[delay]["pkts_per_sec_actual"]))
                report.append(((f'Number of packets sent'), self.report_data[delay]['pkts_sent']))
                report.append(((f'Number of packets received'), self.report_data[delay]['pkts_recieved']))
            self.add_to_stage_report(f'---Stage {self.stage_name} {PassCriteria.to_string(self.stage_pass)}, Running time {datetime.datetime.now() - self.start_time}')
            self.add_to_stage_report(tabulate.tabulate(pd.DataFrame(report), showindex=False))   
            self.add_to_stage_report(f'Stage data saved - {self.csv_path}')
            fig_data = pd.DataFrame.from_dict(self.report_data, orient='index')
            fig_data.reset_index(inplace=True)
            fig_data.rename(columns={'index': 'time_delay'}, inplace=True)
            fig = px.line(fig_data, x='pkts_per_sec_actual', y='percent_of_false_recieved', title='Percentage of packets not recieved by packets per second sent',
              labels={'pkts_per_sec_actual': 'Packets Sent Per Second', 'false_percentage': 'Percentage of packets not recieved'})
            graph_div = fig.to_html(full_html=False, include_plotlyjs='cdn')
            
            # Generate HTML
            table_html = self.template_engine.render_template('table.html', dataframe=self.comparison.to_html(table_id=self.stage_name), table_id=self.stage_name)
            self.report_html = self.template_engine.render_template('stage.html', stage=self, 
                                                                    stage_report=self.report.split('\n'), graph=graph_div)
            
            return self.report
        
        
STAGES = [StressTestStage] 

class StressTest(GenericTest):
    def __init__(self, **kwargs):        
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, test_name=type(self).__name__)
        self.all_messages_in_test = []
        self.stages = [stage(**self.__dict__) for stage in STAGES]
        self.result_indication = INFORMATIVE
        self.test_tooltip = "Stages attempting to stress test a maximum throughput"

    def run(self):
        super().run()
        self.test_pass = PERFECT_SCORE
        for stage in self.stages:
            stage.prepare_stage()
            stage.run()
            self.test_pass = PassCriteria.calc_for_test(self.test_pass, stage)
            self.all_messages_in_test.extend(self.mqttc.get_all_messages_from_topic('data'))
    



import datetime
import os
from wiliot_deployment_tools.api.extended_api import ExtendedEdgeClient
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.gw_certificate.api_if.gw_capabilities import GWCapabilities
from wiliot_deployment_tools.interface.ble_simulator import BLESimulator
from wiliot_deployment_tools.interface.if_defines import SEP
from wiliot_deployment_tools.interface.mqtt import MqttClient

PASS_STATUS = {True: 'PASS', False: 'FAIL'}

# Score values for pass/inconclusive/fail
PERFECT_SCORE = 100
PASS_MINIMUM = 80
INCONCLUSIVE_MINIMUM = 70
INIT_INCONCLUSIVE_MINIMUM = 40
MINIMUM_SCORE = 0

# Results indications for stages. Must always be synced with frontend. 
# 'score' shows as pass/inconclusive/fail. 'info' shows as info/warning.
SCORE_BASED = 'score'
INFORMATIVE = 'info'

class PassCriteria():
    def __init__(self):
        pass
        
    @staticmethod
    def to_string(pass_value:int) -> str:
        if pass_value >= PASS_MINIMUM:
            return 'Pass'
        elif pass_value >= INCONCLUSIVE_MINIMUM:
            return 'Inconclusive'
        else:
            return 'Fail'

    @staticmethod
    def missing_score(pass_value:int) -> int:
        return PERFECT_SCORE - pass_value

    @staticmethod
    def calc_for_stage_uplink(pass_value:int, stage_name:str) -> int:
        error_msg = "Insufficient amount of packets were scanned & uploaded by the gateway"
        return pass_value, error_msg

    @staticmethod
    def calc_for_stage_stress(pass_value: int, stage_name:str) -> int:
        return pass_value
    
    @staticmethod
    def calc_for_stage_downlink(rsquared, slope, stage_name:str):
        error_msg = ''
        if 'Sanity' in stage_name:
            if rsquared > 0:
                return PERFECT_SCORE, error_msg
            else:
                error_msg = 'No advertisements were received from the gateway.'
                return MINIMUM_SCORE, error_msg
        else:
            if rsquared > 0.8 and slope > 0:
                return PERFECT_SCORE, error_msg
            elif rsquared > 0.5 and slope > 0:
                error_msg = "The correlation between 'txMaxDuration' and the board advertisements is suboptimal."
                return INCONCLUSIVE_MINIMUM, error_msg
            else:
                error_msg = "The correlation between 'txMaxDuration' and the board advertisements is weak."
                return MINIMUM_SCORE, error_msg

    @staticmethod
    def calc_for_test(test_pass_value:int, stage) -> int:
        if stage.stage_pass < test_pass_value:
            if 'Geolocation' in stage.stage_name or 'info' in stage.result_indication:
                return test_pass_value
            else:
                return stage.stage_pass
        else:
            return test_pass_value


class GenericTest:
    def __init__(self, mqttc: MqttClient, ble_sim: BLESimulator, 
                 gw_capabilities:GWCapabilities, gw_id, owner_id, test_name, **kwargs):
        # Clients
        self.mqttc = mqttc
        self.ble_sim = ble_sim
        
        # Test-Related
        self.gw_capabilities = gw_capabilities
        self.report = ''
        self.report_html = ''
        self.test_pass = MINIMUM_SCORE
        self.pass_min = PASS_MINIMUM
        self.inconclusive_min = INCONCLUSIVE_MINIMUM
        self.start_time = None
        self.test_name = test_name
        self.test_dir = os.path.join(self.certificate_dir, self.test_name)
        self.env_dirs.create_dir(self.test_dir)
        self.stages = []
        self.test_tooltip = kwargs.get('test_tooltip', 'Missing tooltip')
        self.result_indication = kwargs.get('result_indication', SCORE_BASED)
        
    def __repr__(self):
        return self.test_name
    
    def run(self):
        self.start_time = datetime.datetime.now()
        debug_print(f"Starting Test {self.test_name} : {self.start_time}")
        
    def runtime(self):
        return datetime.datetime.now() - self.start_time
    
    def add_to_test_report(self, report):
        self.report += '\n' + report
    
    def create_test_html(self):
        self.report_html = self.template_engine.render_template('test.html', test=self,
                                                                running_time = self.runtime())
    
    def end_test(self):
        self.create_test_html()

class GenericStage():
    def __init__(self, stage_name, **kwargs):
        #Stage Params
        self.stage_name = stage_name
        self.result_indication = kwargs.get('result_indication', SCORE_BASED)
        self.stage_pass = MINIMUM_SCORE
        self.pass_min = kwargs.get('pass_min', PASS_MINIMUM)
        self.inconclusive_min = INCONCLUSIVE_MINIMUM
        self.report = ''
        self.report_html = ''
        self.start_time = None
        self.csv_path = os.path.join(self.test_dir, f'{self.stage_name}.csv')
        self.stage_tooltip = kwargs.get('stage_tooltip', 'Missing tooltip')
        self.error_summary = ''
        
    def __repr__(self):
        return self.stage_name
    
    def prepare_stage(self):
        debug_print(f'### Starting Stage: {self.stage_name}')

    def add_to_stage_report(self, report):
        self.report += f'{report}\n'
    
    def generate_stage_report(self):
        return self.report
